package dev.fitko.fitconnect.api.services;

import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.EventStatus;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.submission.CreateSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.*;

import java.util.List;
import java.util.UUID;

/**
 * A technical system that creates a submission via the FIT-Connect Submission API.
 * <p>
 * The Sender acts as a common interface wrapping all client functionality for authenticating <p>
 * and creating a valid {@link SubmitSubmission} including encrypted {@link Data} and {@link Attachment}s
 *
 * @see <a href="https://docs.fitko.de/fit-connect/docs/sending/overview">Sending Submissions</a>
 */
public interface Sender {

    /**
     * Validates the public key consisting of the following steps:
     * <p>
     * <ul>
     *     <li>checks if the JSON Web Key is suitable for the encryption</li>
     *     <li>checks if the public key is matching the certificate referenced in the JWK</li>
     *     <li>checks the certificate chain up to the root certificate</li>
     *     <li>checks against a certificate revocation list and/or an OSCP-endpoint with signed response</li>
     * </ul>
     * @param publicKey the public JWK
     *
     * @return {@link ValidationResult} that includes an error if the validation failed
     */
    ValidationResult validatePublicKey(RSAKey publicKey);

    /**
     * Validates the {@link Metadata} structure against a given JSON-schema to ensure its correctness.
     *
     * @param metadata the {@link Metadata} object that is validated
     * @return a {@link ValidationResult}, contains an error if the {@link Metadata} is invalid or doesn't match the schema
     */
    ValidationResult validateMetadata(Metadata metadata);

    /**
     * Checks if a received callback can be trusted by validating the provided request data
     *
     * @param hmac authentication code provided by the callback
     * @param timestamp timestamp provided by the callback
     * @param httpBody HTTP body provided by the callback
     * @param callbackSecret secret owned by the client, which is used to calculate the hmac
     * @return {@code true} if hmac and timestamp provided by the callback meet the required conditions
     */
    ValidationResult validateCallback(String hmac, Long timestamp, String httpBody, String callbackSecret);

    /**
     * Encrypts the submission data payload (json or xml) with JWE (JSON-Web-Encryption).
     *
     * @param publicKey the public encryption key for encryption of {@link Data}
     * @param data      the {@link Data} payload that should be encrypted
     * @return the JWE-encrypted {@link Data}
     *
     * @throws EncryptionException if the encryption fails due to an invalid key or corrupt data
     */
    String encryptBytes(RSAKey publicKey, byte[] data) throws EncryptionException;

    /**
     * Encrypts an object with JWE (JSON-Web-Encryption).
     *
     * @param publicKey the public encryption key
     * @param object  the data payload that should be encrypted
     * @return the JWE-encrypted {@link Data}
     *
     * @throws EncryptionException if the encryption fails due to an invalid key or corrupt data
     */
    String encryptObject(RSAKey publicKey, Object object) throws EncryptionException;

    /**
     * Create message digest (SHA-512) of a given byte[]
     * @param data input bytes that should be hashed
     * @return string representation of hashed byte[]
     */
    String createHash(byte[] data);

    /**
     * Creates and announces a new {@link CreateSubmission}.
     *
     * @param submission with a destinationId, a list of attachmentIds and a serviceType
     * @return the announced submission
     *
     * @throws SubmissionNotCreatedException if the announcement of the submission failed, e.g. due to a technical problem
     */
    SubmissionForPickup createSubmission(CreateSubmission submission) throws SubmissionNotCreatedException;

    /**
     * Uploads encrypted {@link Attachment}s to for given submission id.
     *
     * @param submissionId unique identifier of the  submission
     * @param attachmentId unique identifier of the attachment
     * @param encryptedAttachment string of JWE encrypted attachment
     *
     * @throws AttachmentUploadException if a technical problem occurred upload the attachment
     */
    void uploadAttachment(UUID submissionId, UUID attachmentId, String encryptedAttachment);

    /**
     * Gets {@link Destination} by id.
     *
     * @param destinationId unique identifier of a {@link Destination}
     * @return the destination
     */
    Destination getDestination(UUID destinationId);

    /**
     * Retrieves the public encryption key for a destination.
     *
     * @param destinationId unique identifier of the destination
     *
     * @return the public key for encrypting {@link Metadata}, {@link Data} and {@link Attachment}s
     *
     * @throws KeyNotRetrievedException if a technical problem occurred fetching the key
     */
    RSAKey getEncryptionKeyForDestination(UUID destinationId) throws KeyNotRetrievedException;

    /**
     * Send a {@link SubmitSubmission} that includes encrypted {@link Metadata} and {@link Data}
     *
     * @param submission submission to submit
     * @return sent {@link Submission}
     */
    Submission sendSubmission(SubmitSubmission submission);

    /**
     * Tests if a given string is well-formed JSON syntax.
     *
     * @param json json string that is tested
     *
     * @return a {@link ValidationResult} with an optional error
     */
    ValidationResult validateJsonFormat(String json);

    /**
     * Tests if a given string is well-formed XML syntax.
     *
     * @param xml xml string that is tested
     * @return a {@link ValidationResult} with an optional error
     */
    ValidationResult validateXmlFormat(String xml);

    /**
     * Retrieve the entire event log for a submissions caseId of a specific destination.
     *
     * @param caseId unique identifier of the case the log should be retrieved for
     * @param destinationId unique identifier of the {@link Destination} the log should be retrieved for
     *
     * @return List of {@link EventLogEntry}s for the given case
     */
    List<EventLogEntry> getEventLog(UUID caseId, UUID destinationId);

    /**
     * Retrieve the current status of a {@link Submission}.
     *
     * @param destinationId unique identifier of the {@link Destination} the log should be retrieved for
     * @param caseId unique identifier of the case the log should be retrieved for
     * @param submissionId unique identifier of the submission the log should be retrieved for
     * @param authenticationTags {@link AuthenticationTags} used for SET-Event integrity validation
     *
     * @return {@link EventStatus} the current status
     */
    EventStatus getLastedEvent(UUID destinationId, UUID caseId, UUID submissionId, AuthenticationTags authenticationTags) throws RestApiException;
}
