package dev.fitko.fitconnect.api.services;

import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.EventPayload;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;

import java.util.List;
import java.util.Set;
import java.util.UUID;

/**
 * A technical system that accepts submissions on the administration side.
 *
 * @see <a href="https://docs.fitko.de/fit-connect/docs/receiving/overview">Receiving Submissions</a>
 */
public interface Subscriber {

    /**
     * Decrypts JWE-encrypted string data.
     *
     * @param privateKey       the private key to decrypt the JWE-encrypted string
     * @param encryptedContent JWE encrypted content that should be decrypted
     * @return the decrypted content as byte[]
     */
    byte[] decryptStringContent(RSAKey privateKey, String encryptedContent);

    /**
     * Polls available {@link SubmissionForPickup}s for a given destinationId.
     *
     * @param destinationId restricts the query to a specific destination
     * @param limit         number of submissions in result (max. is 500)
     * @param offset        position in the dataset
     * @return list of found {@link SubmissionForPickup}s
     */
    Set<SubmissionForPickup> pollAvailableSubmissions(UUID destinationId, int offset, int limit);

    /**
     * Gets a {@link Submission}.
     *
     * @param submissionId the unique identifier of a {@link Submission}
     * @return the requested {@link Submission}
     */
    Submission getSubmission(UUID submissionId);

    /**
     * Loads encrypted {@link Attachment} for a {@link Submission}.
     *
     * @param submissionId the unique identifier of a {@link Submission}
     * @param attachmentId unique identifier of the attachments
     * @return encrypted JWE string of attachment
     */
    String fetchAttachment(UUID submissionId, UUID attachmentId);

    /**
     * Retrieve the entire event log for a submissions caseId of a specific destination.
     *
     * @param caseId unique identifier of the case the log should be retrieved for
     * @param destinationId unique identifier of the {@link Destination} the log should be retrieved for
     *
     * @return List of {@link EventLogEntry}s for the given case
     */
    List<EventLogEntry> getEventLog(UUID caseId, UUID destinationId);

    /**
     * Validates the {@link Metadata} structure against a given JSON-schema to ensure its correctness.
     *
     * @param metadata the {@link Metadata} object that is validated
     * @return a {@link ValidationResult}, contains an error if the {@link Metadata} is invalid or doesn't match the schema
     */
    ValidationResult validateMetadata(Metadata metadata);

    /**
     * Validates data integrity of {@link Attachment} or {@link Data}.
     *
     * @param originalHash original hash value of raw data as hex string
     * @param data unencrypted data as byte[]
     * @return a {@link ValidationResult}, contains an error if the hash values are not equal
     */
    ValidationResult validateHashIntegrity(String originalHash, byte[] data);

    /**
     * Checks if a received callback can be trusted by validating the provided request data
     *
     * @param hmac authentication code provided by the callback
     * @param timestamp timestamp provided by the callback
     * @param httpBody HTTP body provided by the callback
     * @param callbackSecret secret owned by the client, which is used to calculate the hmac
     * @return {@code true} if hmac and timestamp provided by the callback meet the required conditions
     */
    ValidationResult validateCallback(String hmac, Long timestamp, String httpBody, String callbackSecret);

    /**
     * Sends a confirmation event if the received submission matches all validations.
     *
     * @param eventPayload contains submissionId, caseId, destination and a list of problems

     * @see <a href="https://docs.fitko.de/fit-connect/docs/receiving/process-and-acknowledge">Process And Acknowledge</a>
     */
    void acceptSubmission(EventPayload eventPayload);

    /**
     * Sends a rejection event if the received submission violates any validation rule.
     *
     * @param eventPayload contains submissionId, caseId, destination and a list of problems

     * @see <a href="https://docs.fitko.de/fit-connect/docs/receiving/process-and-acknowledge">Process And Acknowledge</a>
     */
    void rejectSubmission(EventPayload eventPayload);
}
