package dev.fitko.fitconnect.api.services.crypto;

import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.api.exceptions.DecryptionException;
import dev.fitko.fitconnect.api.exceptions.EncryptionException;

/**
 * A service that allows to encrypt and decrypt {@link Data} and {@link Attachment}s of a {@link SubmitSubmission}
 * via JSON-Web-Encryption.
 *
 * @see <a href="https://datatracker.ietf.org/doc/html/rfc7516">JSON-Web-Encryption</a>
 */
public interface CryptoService {

    /**
     * Decrypts a JWE encrypted string with the given private key.
     *
     * @param privateKey    RSA private key for decryption of JWE
     * @param encryptedData serialized encrypted JWE Hex string
     * @return decrypted JWE string
     *
     * @throws DecryptionException if the payload cannot be decrypted or there was an issue with the key
     */
    String decryptString(RSAKey privateKey, String encryptedData) throws DecryptionException;

    /**
     * Decrypts a JWE encrypted string with the given private key.
     *
     * @param privateKey    RSA private key for decryption of JWE
     * @param encryptedData serialized encrypted JWE Hex string
     * @return decrypted JWE byte array
     *
     * @throws DecryptionException if the payload cannot be decrypted or there was an issue with the key
     */
    byte[] decryptBytes(RSAKey privateKey, String encryptedData) throws DecryptionException;

    /**
     * Encrypts a string with the given public key.
     *
     * @param encryptionKey RSA public key for encryption of string payload
     * @param data string data that should be encrypted
     * @return Hex string of the encrypted JWE object
     *
     * @throws EncryptionException if the payload cannot be encrypted or there was an issue with the key
     */
    String encryptString(RSAKey encryptionKey, String data) throws EncryptionException;

    /**
     * Encrypts an object with the given public key.
     *
     * @param encryptionKey RSA public key for encryption of string payload
     * @param obj   object  that should be encrypted
     * @return Hex string of the encrypted JWE object
     *
     * @throws EncryptionException if the payload cannot be encrypted or there was an issue with the key
     */
    String encryptObject(RSAKey encryptionKey, Object obj);

    /**
     * Encrypts a byte[] payload with the given public key.
     *
     * @param encryptionKey RSA public key the payload is encrypted with
     * @param bytes     byte[] of the data that should be encrypted
     * @return Hex string of the encrypted JWE object
     *
     * @throws EncryptionException if the payload cannot be encrypted or there was an issue with the key
     */
    String encryptBytes(RSAKey encryptionKey, byte[] bytes) throws EncryptionException;

    /**
     * Creates a message digest hash from the given input byte[]
     *
     * @param data byte[] that should be hashed
     * @return hex encoded string of the hashed data
     */
    String hashBytes(byte[] data);

}
