package dev.fitko.fitconnect.api.services.events;

import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.event.EventLog;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.EventStatus;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.exceptions.EventLogException;
import dev.fitko.fitconnect.api.exceptions.RestApiException;

import java.util.List;
import java.util.UUID;

/**
 * A service that provides access to the {@link Submission}s {@link EventLog} via the FIT-Connect REST-API.
 * Every event is a recorded step within the process of sending a submission or a case.
 * Events are encoded as security event tokens (SET).
 *
 * @see <a href="https://docs.fitko.de/fit-connect/docs/getting-started/event-log/events">Event Documentation</a>
 */
public interface EventLogService {

    /**
     * Get the {@link EventLog} by caseId.
     *
     * @param caseId unique case identifier
     * @param destinationId unique identifier of the destination
     * @return list of {@link EventLogEntry}
     */
    List<EventLogEntry> getEventLog(UUID caseId, UUID destinationId) throws RestApiException, EventLogException;

    /**
     * Retrieve the current status of a {@link Submission}.
     *
     * @param destinationId unique identifier of the {@link Destination} the log should be retrieved for
     * @param caseId unique identifier of the case the log should be retrieved for
     * @param submissionId unique identifier of the submission the log should be retrieved for
     * @param authenticationTags {@link AuthenticationTags} used for SET-Event integrity validation
     *
     * @return {@link EventStatus} the current status
     */
    EventStatus getLastedEvent(UUID destinationId, UUID caseId, UUID submissionId, AuthenticationTags authenticationTags) throws RestApiException, EventLogException;

    /**
     * Send an event for a given caseId.
     *
     * @param caseId unique case identifier
     * @param signedAndSerializedSET the serialised and signed event as SET string
     */
    void sendEvent(UUID caseId, String signedAndSerializedSET) throws RestApiException;
}
