package dev.fitko.fitconnect.api.services.validation;

import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;

import java.util.Map;

/**
 * Validator for publicKeys and metadata.
 *
 * @see <a href="https://docs.fitko.de/fit-connect/docs/sending/encrypt/#certificateValidation">Certificate Validation</a>
 */
public interface ValidationService {

    /**
     * Validates the public key consisting of the following steps:
     * <p>
     * <ul>
     *     <li>checks if the JSON Web Key has key operation `wrap_key`</li>
     *     <li>checks if the JSON Web Key is suitable for the encryption</li>
     *     <li>checks if the public key is matching the certificate referenced in the JWK</li>
     *     <li>checks the certificate chain up to the root certificate</li>
     *     <li>checks against a certificate revocation list and/or an OSCP-endpoint with signed response</li>
     * </ul>
     * @param publicKey the public JWK
     *
     * @return {@link ValidationResult} that includes an error if the validation failed
     */
    ValidationResult validateEncryptionPublicKey(RSAKey publicKey);

    /**
     * Validates the public signature key with key-operation `verify`
     *
     * @param signatureKey the public signature JWK
     * @return {@link ValidationResult} that includes an error if the validation failed
     */
    ValidationResult validateSignaturePublicKey(RSAKey signatureKey);

    /**
     * Validates the metadata against a given schema.
     *
     * @param metadata the metadata object that should be validated
     *
     * @return a {@link ValidationResult} with an optional error
     */
    ValidationResult validateMetadataSchema(Metadata metadata);

    /**
     * Validates a set event against a given schema.
     *
     * @param setEventPayload the set event that should be validated
     *
     * @return a {@link ValidationResult} with an optional error
     */
    ValidationResult validateSetEventSchema(String setEventPayload);

    /**
     * Validates signature payload claims against a given schema.
     *
     * @param destinationPayload the payload to be validated
     *
     * @return a {@link ValidationResult} with an optional error
     */
    ValidationResult validateDestinationSchema(Map<String, Object> destinationPayload);

    /**
     * Compares a given byte[] to its original hash value.
     *
     * @param originalHexHash hash of the data as hex string
     * @param data byte[] of data that will be compared to the original hash
     * @return a {@link ValidationResult} with an optional error
     */
    ValidationResult validateHashIntegrity(String originalHexHash, byte[] data);

    /**
     * Tests if a given string is well-formed JSON syntax.
     *
     * @param json json string that is tested
     *
     * @return a {@link ValidationResult} with an optional error
     */
    ValidationResult validateJsonFormat(String json);

    /**
     * Tests if a given string is well-formed XML syntax.
     *
     * @param xml xml string that is tested
     *
     * @return a {@link ValidationResult} with an optional error
     */
    ValidationResult validateXmlFormat(String xml);

    /**
     * Checks if a received callback can be trusted by validating the provided request data
     *
     * @param hmac authentication code provided by the callback
     * @param timestamp timestamp provided by the callback
     * @param httpBody HTTP body provided by the callback
     * @param callbackSecret secret owned by the client, which is used to calculate the hmac
     * @return {@code true} if hmac and timestamp provided by the callback meet the required conditions
     */
    ValidationResult validateCallback(String hmac, Long timestamp, String httpBody, String callbackSecret);
}
