package dev.fitko.fitconnect.client;

import dev.fitko.fitconnect.api.domain.model.route.Area;
import dev.fitko.fitconnect.api.domain.model.route.Route;
import dev.fitko.fitconnect.api.domain.model.route.RouteResult;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import dev.fitko.fitconnect.api.exceptions.RoutingException;
import dev.fitko.fitconnect.api.services.routing.RoutingService;
import dev.fitko.fitconnect.client.router.DestinationSearch;
import dev.fitko.fitconnect.core.routing.RouteVerifier;

import java.util.List;

public final class RoutingClient {

    private final RoutingService routingService;
    private final RouteVerifier routeVerifier;

    public RoutingClient(final RoutingService routingService, final RouteVerifier routeVerifier) {
        this.routingService = routingService;
        this.routeVerifier = routeVerifier;
    }

    /**
     * Finds a list of {@link Route}s by a given service identifier and an area search criterion.
     *
     * @param search search parameters that contain leikaKey and one other area search criterion see {@link DestinationSearch}
     * @return list of found {@link Route}s matching the search criteria
     * @throws RoutingException if the signature validation failed
     * @throws RestApiException if a technical error occurred during the query
     */

    public List<Route> findDestinations(final DestinationSearch search) throws RoutingException, RestApiException {
        final RouteResult routeResult = routingService.getRoutes(search.getLeikaKey(), search.getArs(), search.getAgs(), search.getAreaId(), search.getOffset(), search.getLimit());
        final ValidationResult result = routeVerifier.validateRouteDestinations(routeResult.getRoutes(), search.getLeikaKey(), search.getArs());
        if (result.hasError()) {
            throw new RoutingException(result.getError().getMessage(), result.getError());
        }
        return routeResult.getRoutes();
    }

    /**
     * Finds a list of {@link Area}s based on a filter criterion, e.g. a zip code or city.
     *
     * @param filter filter criterion as string
     * @param offset offset to start from
     * @param limit  max entries
     * @return list of {@link Area}
     * @throws RestApiException if a technical error occurred during the query
     */
    public List<Area> findAreas(final String filter, final int offset, final int limit) throws RestApiException {
        return findAreas(List.of(filter), offset, limit);
    }

    /**
     * Find a list {@link Area}s based on a list of multiple filter criteria, e.g. a zip code or a city as in List.of("04229", "Leip*").
     *
     * @param filters list of string filters
     * @param offset  offset to start from
     * @param limit   max entries
     * @return list of {@link Area}
     * @throws RestApiException if a technical error occurred during the query
     */
    public List<Area> findAreas(final List<String> filters, final int offset, final int limit) throws RestApiException {
        return routingService.getAreas(filters, offset, limit).getAreas();
    }
}
