package dev.fitko.fitconnect.client;

import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.EventStatus;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.submission.SentSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.client.sender.model.EncryptedSubmissionPayload;
import dev.fitko.fitconnect.client.sender.model.SubmissionPayload;
import dev.fitko.fitconnect.client.sender.strategies.SendEncryptedSubmissionStrategy;
import dev.fitko.fitconnect.client.sender.strategies.SendNewSubmissionStrategy;
import dev.fitko.fitconnect.client.util.ValidDataGuard;

import java.util.List;
import java.util.Optional;
import java.util.UUID;

/**
 * A sender-side client for announcing and handing in a {@link SubmitSubmission}
 */
public class SenderClient {

    private final Sender sender;

    private final ValidDataGuard dataGuard;

    public SenderClient(final Sender sender) {
        this.sender = sender;
        dataGuard = new ValidDataGuard(sender);
    }

    /**
     * Retrieve the public encryption key
     *
     * @param destinationId unique identifier of a {@link Destination}
     * @return optional string containing the public JWK
     */
    public Optional<String> getPublicKey(final UUID destinationId) {
        final Optional<RSAKey> key = Optional.ofNullable(sender.getEncryptionKeyForDestination(destinationId));
        return key.isEmpty() ? Optional.empty() : Optional.of(key.get().toJSONString());
    }

    /**
     * Retrieve the entire event log for a submissions caseId of a specific destination.
     *
     * @param caseId        unique identifier of the case the log should be retrieved for
     * @param destinationId unique identifier of the {@link Destination} the log should be retrieved for
     * @return List of {@link EventLogEntry}s for the given case
     */
    public List<EventLogEntry> getEventLog(final UUID caseId, final UUID destinationId) {
        return sender.getEventLog(caseId, destinationId);
    }

    /**
     * Retrieve the current status of a {@link Submission}.
     *
     * @param sentSubmission the original {@link SentSubmission} the status should be retrieved for
     * @return {@link EventStatus} the current status
     */
    public EventStatus getStatusForSubmission(final SentSubmission sentSubmission) {
        final UUID destinationId = sentSubmission.getDestinationId();
        final UUID caseId = sentSubmission.getCaseId();
        final UUID submissionId = sentSubmission.getSubmissionId();
        final AuthenticationTags authenticationTags = sentSubmission.getAuthenticationTags();
        return sender.getLastedEvent(destinationId, caseId, submissionId, authenticationTags);
    }

    /**
     * Checks if a received callback can be trusted by validating the provided request data.
     *
     * @param hmac           authentication code provided by the callback
     * @param timestamp      timestamp provided by the callback
     * @param httpBody       HTTP body provided by the callback
     * @param callbackSecret secret owned by the client, which is used to calculate the hmac
     * @return {@code ValidationResult.ok()} if hmac and timestamp provided by the callback meet the required conditions
     */
    public ValidationResult validateCallback(final String hmac, final Long timestamp, final String httpBody, final String callbackSecret) {
        return sender.validateCallback(hmac, timestamp, httpBody, callbackSecret);
    }

    /**
     * Send submission with attachments and data to FIT-Connect API. Encryption is handled by the SDK.
     *
     * @return {@link SentSubmission} object of the handed in submission
     */
    public SentSubmission submit(final SubmissionPayload submissionPayload) {
        dataGuard.ensureValidDataPayload(submissionPayload);
        return new SendNewSubmissionStrategy(sender).send(submissionPayload);
    }

    /**
     * Send already encrypted submission to FIT-Connect API.
     *
     * @return {@link SentSubmission} object of the handed in submission
     */
    public SentSubmission submit(final EncryptedSubmissionPayload encryptedSubmissionPayload) {
        dataGuard.ensureValidDataPayload(encryptedSubmissionPayload);
        return new SendEncryptedSubmissionStrategy(sender).send(encryptedSubmissionPayload);
    }
}
