package dev.fitko.fitconnect.client;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.DataIntegrityException;
import dev.fitko.fitconnect.api.exceptions.DecryptionException;
import dev.fitko.fitconnect.api.exceptions.EventCreationException;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import dev.fitko.fitconnect.api.services.Subscriber;
import dev.fitko.fitconnect.client.subscriber.ReceivedSubmission;
import dev.fitko.fitconnect.client.subscriber.model.DecryptedAttachmentPayload;
import dev.fitko.fitconnect.client.subscriber.model.ReceivedAttachment;
import dev.fitko.fitconnect.client.subscriber.model.ReceivedData;
import dev.fitko.fitconnect.core.util.StopWatch;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import java.util.UUID;
import java.util.function.Function;
import java.util.stream.Collectors;

import static com.fasterxml.jackson.databind.DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES;

/**
 * A subscriber-side client for retrieving submissions.
 */
public class SubscriberClient {

    private static final Logger LOGGER = LoggerFactory.getLogger(SubscriberClient.class);
    private static final ObjectMapper MAPPER = new ObjectMapper().configure(FAIL_ON_UNKNOWN_PROPERTIES, false);
    private static final int DEFAULT_SUBMISSION_LIMIT = 500;
    private final Subscriber subscriber;
    private final RSAKey privateKey;

    public SubscriberClient(final Subscriber subscriber, final RSAKey privateKey) {
        this.subscriber = subscriber;
        this.privateKey = privateKey;
    }

    /**
     * Loads a list of available {@link SubmissionForPickup} that were submitted to the {@link Subscriber}.
     *
     * @param destinationId unique identifier for a destination
     * @return set of the first 0..500 available submissions
     */
    public Set<SubmissionForPickup> getAvailableSubmissions(final UUID destinationId) {
        return getAvailableSubmissions(destinationId, 0, DEFAULT_SUBMISSION_LIMIT);
    }

    /**
     * Loads a list of available {@link SubmissionForPickup} that were submitted to the {@link Subscriber}.
     *
     * @param destinationId unique identifier for a destination
     * @param offset        position in the dataset
     * @param limit         number of submissions in result (max. is 500)
     * @return set of available submissions in the given range
     */
    public Set<SubmissionForPickup> getAvailableSubmissions(final UUID destinationId, final int offset, final int limit) {
        final Set<SubmissionForPickup> submissions = subscriber.pollAvailableSubmissions(destinationId, offset, limit);
        LOGGER.info("Received {} submission(s) for destination {}", submissions.size(), destinationId);
        return submissions;
    }

    /**
     * Retrieve the entire event log for a submissions caseId of a specific destination.
     *
     * @param caseId        unique identifier of the case the log should be retrieved for
     * @param destinationId unique identifier of the {@link Destination} the log should be retrieved for
     * @return List of {@link EventLogEntry}s for the given case
     */
    public List<EventLogEntry> getEventLog(final UUID caseId, final UUID destinationId) {
        return subscriber.getEventLog(caseId, destinationId);
    }

    /**
     * Loads a single {@link SubmissionForPickup} by id.
     *
     * @param submissionId unique identifier of a {@link SubmissionForPickup}
     * @return {@link ReceivedSubmission} to get the submission {@link Metadata}, {@link Data} and {@link Attachment}s as well as
     * accept or reject the loaded submission
     */
    public ReceivedSubmission requestSubmission(final UUID submissionId) {

        try {

            final var startTimeDownloadSubmission = StopWatch.start();
            final Submission submission = subscriber.getSubmission(submissionId);
            LOGGER.info("Downloading submission took {}", StopWatch.stopWithFormattedTime(startTimeDownloadSubmission));

            LOGGER.info("Decrypting metadata ...");
            final Metadata metadata = decryptMetadata(submission.getEncryptedMetadata());
            final ValidationResult metadataValidation = subscriber.validateMetadata(metadata);
            if (metadataValidation.hasError()) {
                LOGGER.error("Metadata does not match schema", metadataValidation.getError());
                return null;
            }

            LOGGER.info("Decrypting data ...");
            final byte[] decryptedData = decryptData(submission.getEncryptedData());
            final String hashFromSender = getDataHashFromMetadata(metadata);
            final ValidationResult dataValidation = subscriber.validateHashIntegrity(hashFromSender, decryptedData);
            if (dataValidation.hasError()) {
                LOGGER.error("Data might be corrupted, hash-sum does not match", dataValidation.getError());
                return null;
            }

            LOGGER.info("Loading and decrypting attachments ...");
            final List<Attachment> attachmentMetadata = metadata.getContentStructure().getAttachments();
            final List<DecryptedAttachmentPayload> decryptedAttachmentPayloads = loadAttachments(submissionId, attachmentMetadata);
            final ValidationResult attachmentValidation = validateAttachments(decryptedAttachmentPayloads);
            if (attachmentValidation.hasError()) {
                LOGGER.error("One of the attachments is invalid", attachmentValidation.getError());
                return null;
            }

            return buildReceivedSubmission(submission, metadata, decryptedData, decryptedAttachmentPayloads);

        } catch (final DecryptionException e) {
            LOGGER.error("Decrypting payload failed", e);
        } catch (final RestApiException e) {
            LOGGER.error("API request failed", e);
        } catch (final EventCreationException e) {
            LOGGER.error("Creating SET event failed", e);
        } catch (final IOException e) {
            LOGGER.error("Reading metadata failed", e);
        }

        return null;
    }

    public ValidationResult validateCallback(final String hmac, final Long timestamp, final String httpBody, final String callbackSecret) {
        return subscriber.validateCallback(hmac, timestamp, httpBody, callbackSecret);
    }

    private ReceivedSubmission buildReceivedSubmission(final Submission submission, final Metadata metadata, final byte[] decryptedData, final List<DecryptedAttachmentPayload> attachments) {
        final MimeType mimeType = metadata.getContentStructure().getData().getSubmissionSchema().getMimeType();
        final ReceivedData receivedData = new ReceivedData(new String(decryptedData, StandardCharsets.UTF_8), mimeType);
        final List<ReceivedAttachment> receivedAttachments = attachments.stream().map(mapToReceivedAttachment()).collect(Collectors.toList());
        return new ReceivedSubmission(subscriber, submission, metadata, receivedData, receivedAttachments);
    }

    private static Function<DecryptedAttachmentPayload, ReceivedAttachment> mapToReceivedAttachment() {
        return payload -> ReceivedAttachment.builder()
                .attachmentId(payload.getAttachmentMetadata().getAttachmentId())
                .filename(payload.getAttachmentMetadata().getFilename())
                .mimeType(payload.getAttachmentMetadata().getMimeType())
                .description(payload.getAttachmentMetadata().getDescription())
                .data(payload.getDecryptedContent())
                .build();
    }

    private List<DecryptedAttachmentPayload> loadAttachments(final UUID submissionId, final List<Attachment> attachmentMetadata) {
        if (attachmentMetadata == null || attachmentMetadata.isEmpty()) {
            LOGGER.info("Submission contains no attachments");
            return Collections.emptyList();
        }
        final List<DecryptedAttachmentPayload> receivedAttachments = new ArrayList<>();
        for (final Attachment metadata : attachmentMetadata) {
            final String encryptedAttachment = downloadAttachment(submissionId, metadata);
            final byte[] decryptedAttachment = decryptAttachment(metadata, encryptedAttachment);
            final DecryptedAttachmentPayload decryptedAttachmentPayload = DecryptedAttachmentPayload.builder()
                    .decryptedContent(decryptedAttachment)
                    .attachmentMetadata(metadata)
                    .build();
            receivedAttachments.add(decryptedAttachmentPayload);
        }
        return receivedAttachments;
    }

    private byte[] decryptAttachment(final Attachment metadata, final String encryptedAttachment) {
        final var startDecryption = StopWatch.start();
        final byte[] decryptedAttachment = subscriber.decryptStringContent(privateKey, encryptedAttachment);
        LOGGER.info("Decrypting attachment {} took {}", metadata.getAttachmentId(), StopWatch.stopWithFormattedTime(startDecryption));
        return decryptedAttachment;
    }

    private String downloadAttachment(final UUID submissionId, final Attachment metadata) {
        final var startDownload = StopWatch.start();
        final String encryptedAttachment = subscriber.fetchAttachment(submissionId, metadata.getAttachmentId());
        LOGGER.info("Downloading attachment {} took {}", metadata.getAttachmentId(), StopWatch.stopWithFormattedTime(startDownload));
        return encryptedAttachment;
    }

    private byte[] decryptData(final String encryptedData) {
        return subscriber.decryptStringContent(privateKey, encryptedData);
    }

    private Metadata decryptMetadata(final String encryptedMetadata) throws IOException {
        final byte[] metadataBytes = subscriber.decryptStringContent(privateKey, encryptedMetadata);
        return MAPPER.readValue(metadataBytes, Metadata.class);
    }

    private ValidationResult validateAttachments(final List<DecryptedAttachmentPayload> decryptedAttachmentPayloads) {
        for (final DecryptedAttachmentPayload decryptedAttachment : decryptedAttachmentPayloads) {
            final Attachment attachmentMetadata = decryptedAttachment.getAttachmentMetadata();
            final UUID attachmentId = attachmentMetadata.getAttachmentId();
            final ValidationResult result = subscriber.validateHashIntegrity(attachmentMetadata.getHash().getContent(), decryptedAttachment.getDecryptedContent());
            if (result.hasError()) {
                LOGGER.error("Attachment data for id {} is corrupted", attachmentId, result.getError());
                return ValidationResult.error(new DataIntegrityException("Attachment " + attachmentId + " is corrupt"));
            } else {
                LOGGER.info("Attachment {} with id {} is valid", attachmentMetadata.getFilename(), attachmentId);
            }
        }
        return ValidationResult.ok();
    }

    private String getDataHashFromMetadata(final Metadata metadata) {
        return metadata.getContentStructure()
                .getData()
                .getHash()
                .getContent();
    }
}
