package dev.fitko.fitconnect.client.factory;

import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.config.BuildInfo;
import dev.fitko.fitconnect.api.exceptions.InitializationException;
import org.yaml.snakeyaml.Yaml;

import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Files;
import java.nio.file.Path;

public final class ApplicationConfigLoader {

    private static final String BUILD_INFO_PATH = "buildinfo.yaml";

    private ApplicationConfigLoader() {
    }

    /**
     * Load ApplicationConfig from path.
     *
     * @param configPath path to config file
     * @return ApplicationConfig
     * @throws InitializationException if the config file could not be loaded
     */
    public static ApplicationConfig loadConfig(final Path configPath) {
        try {
            return loadConfigFromYaml(Files.readString(configPath));
        } catch (final IOException | NullPointerException e) {
            throw new InitializationException(e.getMessage(), e);
        }
    }

    /**
     * Load ApplicationConfig from path string.
     *
     * @param configPath string of path to config file
     * @return ApplicationConfig
     * @throws InitializationException if the config file could not be loaded
     */
    public static ApplicationConfig loadConfig(final String configPath) {
        return loadConfig(Path.of(configPath));
    }

    /**
     * Load ApplicationConfig from yaml string.
     *
     * @param configYaml string content of the yaml config file
     * @return ApplicationConfig
     */
    public static ApplicationConfig loadConfigFromYaml(final String configYaml) {
        return new Yaml().loadAs(configYaml, ApplicationConfig.class);
    }

    /**
     * Load BuildInfo properties.
     *
     * @return {@link BuildInfo}
     */
    public static BuildInfo loadBuildInfo() {
        final InputStream resource = ApplicationConfigLoader.class.getClassLoader().getResourceAsStream(BUILD_INFO_PATH);
        return new Yaml().loadAs(resource, BuildInfo.class);
    }
}
