package dev.fitko.fitconnect.client.sender.strategies;

import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.destination.DestinationService;
import dev.fitko.fitconnect.api.domain.model.metadata.ContentStructure;
import dev.fitko.fitconnect.api.domain.model.metadata.Hash;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.PublicServiceType;
import dev.fitko.fitconnect.api.domain.model.metadata.SignatureType;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Purpose;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.submission.CreateSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.SentSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.ServiceType;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.AttachmentCreationException;
import dev.fitko.fitconnect.api.exceptions.EncryptionException;
import dev.fitko.fitconnect.api.exceptions.InvalidKeyException;
import dev.fitko.fitconnect.api.exceptions.KeyNotRetrievedException;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import dev.fitko.fitconnect.api.exceptions.SchemaNotFoundException;
import dev.fitko.fitconnect.api.exceptions.SubmissionNotCreatedException;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.client.sender.model.AttachmentPayload;
import dev.fitko.fitconnect.client.sender.model.SubmissionPayload;
import dev.fitko.fitconnect.core.util.StopWatch;
import org.apache.tika.Tika;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.net.URLConnection;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.Collection;
import java.util.List;
import java.util.Objects;
import java.util.UUID;

import java.util.stream.Collectors;

import static dev.fitko.fitconnect.api.config.SchemaConfig.METADATA_V_1_0_0;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildCreateSubmission;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildSentSubmission;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildSubmitSubmission;


public class SendNewSubmissionStrategy {

    private static final Logger LOGGER = LoggerFactory.getLogger(SendNewSubmissionStrategy.class);

    private final Sender sender;

    public SendNewSubmissionStrategy(final Sender sender) {
        this.sender = sender;
    }

    public SentSubmission send(final SubmissionPayload submissionPayload) {

        final UUID destinationId = submissionPayload.getDestinationId();
        final ServiceType serviceType = submissionPayload.getServiceType();

        try {

            final RSAKey encryptionKey = sender.getEncryptionKeyForDestination(destinationId);
            final Destination destination = sender.getDestination(destinationId);

            final List<AttachmentPayload> encryptedAttachments = encryptAndHashAttachments(encryptionKey, submissionPayload.getAttachments());
            final CreateSubmission newSubmission = buildCreateSubmission(destinationId, encryptedAttachments, serviceType);

            final SubmissionForPickup announcedSubmission = sender.createSubmission(newSubmission);
            final UUID announcedSubmissionId = announcedSubmission.getSubmissionId();

            final var startTimeAttachmentUpload = StopWatch.start();
            uploadAttachments(encryptedAttachments, announcedSubmissionId);
            LOGGER.info("Uploading attachments took {}", StopWatch.stopWithFormattedTime(startTimeAttachmentUpload));

            LOGGER.info("Creating metadata");
            final Metadata metadata = buildMetadata(submissionPayload, destination, encryptedAttachments);

            final ValidationResult validatedMetadata = sender.validateMetadata(metadata);
            if (validatedMetadata.hasError()) {
                LOGGER.error("Metadata does not match schema", validatedMetadata.getError());
                return null;
            }

            final String encryptedData = sender.encryptBytes(encryptionKey, submissionPayload.getData().getBytes(StandardCharsets.UTF_8));
            final String encryptedMetadata = sender.encryptObject(encryptionKey, metadata);

            final var startTimeSubmissionUpload = StopWatch.start();
            final Submission submission = sender.sendSubmission(buildSubmitSubmission(announcedSubmissionId, encryptedData, encryptedMetadata));
            LOGGER.info("Uploading submission took {}", StopWatch.stopWithFormattedTime(startTimeSubmissionUpload));

            LOGGER.info("SUCCESSFULLY HANDED IN SUBMISSION !");
            return buildSentSubmission(encryptedAttachments, encryptedData, encryptedMetadata, submission);

        } catch (final EncryptionException e) {
            LOGGER.error("Encrypting submission failed", e);
        } catch (final RestApiException e) {
            LOGGER.error("Sending submission failed", e);
        } catch (final SchemaNotFoundException e) {
            LOGGER.error("Required schema to send valid submission not found", e);
        } catch (final SubmissionNotCreatedException e) {
            LOGGER.error("Failed to announce new submission", e);
        } catch (final KeyNotRetrievedException e) {
            LOGGER.error("Getting encryption key for destination {} failed", destinationId, e);
        } catch (final InvalidKeyException e) {
            LOGGER.error("Key for destination {} is invalid", destinationId, e);
        } catch (final AttachmentCreationException e) {
            LOGGER.error("Reading file failed. Attachment will not be created.", e);
        }

        return null;
    }

    private Metadata buildMetadata(final SubmissionPayload submissionPayload, final Destination destination, final List<AttachmentPayload> encryptedAttachments) {
        final String hashedData = sender.createHash(submissionPayload.getData().getBytes(StandardCharsets.UTF_8));

        final MimeType dataMimeType = submissionPayload.getDataMimeType();
        final URI schemaUri = getSubmissionSchemaFromDestination(destination, dataMimeType);
        final Data data = createData(dataMimeType, hashedData, schemaUri);
        final PublicServiceType publicServiceType = buildPublicServiceType(submissionPayload.getServiceType());
        final List<Attachment> attachmentMetadata = encryptedAttachments.stream().map(this::toHashedAttachment).collect(Collectors.toList());

        final var contentStructure = new ContentStructure();
        contentStructure.setAttachments(attachmentMetadata);
        contentStructure.setData(data);

        final var metadata = new Metadata();
        metadata.setSchema(METADATA_V_1_0_0.toString());
        metadata.setContentStructure(contentStructure);
        metadata.setPublicServiceType(publicServiceType);

        return metadata;
    }

    private Data createData(final MimeType mimeType, final String hashedData, final URI schemaUri) {
        final var hash = new Hash();
        hash.setContent(hashedData);
        hash.setSignatureType(SignatureType.SHA_512);

        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setMimeType(mimeType);
        submissionSchema.setSchemaUri(schemaUri);

        final var data = new Data();
        data.setSubmissionSchema(submissionSchema);
        data.setHash(hash);
        return data;
    }

    private PublicServiceType buildPublicServiceType(final ServiceType serviceType) {
        final var publicServiceType = new PublicServiceType();
        publicServiceType.setIdentifier(serviceType.getIdentifier());
        if (serviceType.getName() != null) {
            publicServiceType.setName(serviceType.getName());
        }
        if (serviceType.getDescription() != null) {
            publicServiceType.setDescription(serviceType.getDescription());
        }
        return publicServiceType;
    }

    private Attachment toHashedAttachment(final AttachmentPayload attachmentPayload) {
        final var attachment = new Attachment();
        attachment.setAttachmentId(attachmentPayload.getAttachmentId());
        attachment.setPurpose(Purpose.ATTACHMENT);
        attachment.setFilename(attachmentPayload.getFile().getName());
        attachment.setMimeType(attachmentPayload.getMimeType());

        final var hash = new Hash();
        hash.setContent(attachmentPayload.getHashedData());
        hash.setSignatureType(SignatureType.SHA_512);
        attachment.setHash(hash);
        return attachment;
    }

    private URI getSubmissionSchemaFromDestination(final Destination destination, final MimeType mimeType) {
        return destination.getServices().stream()
                .map(DestinationService::getSubmissionSchemas)
                .flatMap(Collection::stream)
                .filter(schema -> schema.getMimeType().equals(mimeType))
                .map(SubmissionSchema::getSchemaUri)
                .findFirst()
                .orElseThrow(() -> new SchemaNotFoundException("Destination does not support data with mime-type " + mimeType));
    }

    private void uploadAttachments(final List<AttachmentPayload> attachmentPayloads, final UUID submissionId) {
        if (attachmentPayloads.isEmpty()) {
            LOGGER.info("No attachments to upload");
        } else {
            LOGGER.info("Uploading {} attachment(s)", attachmentPayloads.size());
            attachmentPayloads.forEach(a -> sender.uploadAttachment(submissionId, a.getAttachmentId(), a.getEncryptedData()));
        }
    }

    private List<AttachmentPayload> encryptAndHashAttachments(final RSAKey encryptionKey, final List<File> attachments) {
        return attachments.stream()
                .map(file -> AttachmentPayload.builder().file(file).attachmentId(UUID.randomUUID()).build())
                .filter(Objects::nonNull)
                .map(payload -> encryptAndHashAttachment(encryptionKey, payload))
                .collect(Collectors.toList());
    }

    private AttachmentPayload encryptAndHashAttachment(final RSAKey encryptionKey, final AttachmentPayload attachmentPayload) {
        final File file = attachmentPayload.getFile();
        try {
            final byte[] rawData = Files.readAllBytes(Paths.get(file.getPath()));
            final String encryptedAttachment = sender.encryptBytes(encryptionKey, rawData);
            final String hashedBytes = sender.createHash(rawData);
            return attachmentPayload.withEncryptedData(encryptedAttachment)
                    .withHashedData(hashedBytes)
                    .withMimeType(detectMimeTypeFromFile(attachmentPayload));
        } catch (final Exception e) {
            throw new AttachmentCreationException("Attachment '" + file.getAbsolutePath() + "' could not be created ", e);
        }
    }

    private String detectMimeTypeFromFile(final AttachmentPayload attachmentPayload) {
        final File file = attachmentPayload.getFile();
        String mimeType;
        try {
            final Tika mimeTypeGuesser = new Tika();
            mimeType = mimeTypeGuesser.detect(file);
        } catch (final IOException e) {
            mimeType = URLConnection.guessContentTypeFromName(file.getName());
        }
        LOGGER.info("Detected attachment mime-type {}", mimeType);
        return mimeType;
    }
}
