package dev.fitko.fitconnect.client.subscriber;

import dev.fitko.fitconnect.api.domain.model.event.Event;
import dev.fitko.fitconnect.api.domain.model.event.EventPayload;
import dev.fitko.fitconnect.api.domain.model.event.problems.Problem;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.services.Subscriber;
import dev.fitko.fitconnect.client.subscriber.model.ReceivedAttachment;
import dev.fitko.fitconnect.client.subscriber.model.ReceivedData;

import java.net.URI;
import java.util.List;
import java.util.UUID;

public class ReceivedSubmission {

    private final transient Subscriber subscriber;
    private final Submission submission;
    private final Metadata metadata;
    private final ReceivedData receivedData;
    private final List<ReceivedAttachment> receivedAttachments;

    public ReceivedSubmission(final Subscriber subscriber, final Submission submission, final Metadata metadata, final ReceivedData receivedData, final List<ReceivedAttachment> receivedAttachments) {
        this.subscriber = subscriber;
        this.submission = submission;
        this.metadata = metadata;
        this.receivedData = receivedData;
        this.receivedAttachments = receivedAttachments;
    }

    /**
     * Send an accept-submission {@link Event} to confirm that the handed in submission was accepted.
     *
     * @see <a href="https://docs.fitko.de/fit-connect/docs/getting-started/event-log/events">FIT-Connect Events</a>
     */
    public void acceptSubmission() {
        subscriber.acceptSubmission(new EventPayload(submission));
    }

    /**
     * Send a reject-submission {@link Event} to confirm that the handed in submission was rejected.
     *
     * @param rejectionProblems list of {@link Problem}s that give more detailed information on why the submission was rejected
     * @see <a href="https://docs.fitko.de/fit-connect/docs/getting-started/event-log/events">FIT-Connect Events</a>
     */
    public void rejectSubmission(final List<Problem> rejectionProblems) {
        subscriber.rejectSubmission(new EventPayload(submission, rejectionProblems));
    }

    /**
     * Access decrypted data.
     *
     * @return data as string
     */
    public String getData() {
        return receivedData.getData();
    }

    /**
     * Get data's mime-type.
     *
     * @return {@link MimeType}
     */
    public MimeType getDataMimeType() {
        return receivedData.getMimeType();
    }

    /**
     * Get the data's schema reference.
     *
     * @return data schema as URI
     */
    public URI getDataSchemaUri() {
        return metadata.getContentStructure().getData().getSubmissionSchema().getSchemaUri();
    }

    /**
     * Access list of decrypted attachments.
     *
     * @return list of {@link ReceivedAttachment}
     */
    public List<ReceivedAttachment> getAttachments() {
        return receivedAttachments;
    }

    /**
     * Get the submissions {@link Metadata}.
     *
     * @return metadata
     */
    public Metadata getSubmissionMetadata() {
        return metadata;
    }

    /**
     * Get the submissions caseId.
     *
     * @return caseId as UUID
     */
    public UUID getCaseId() {
        return submission.getCaseId();
    }

    /**
     * Get the submissions destinationId.
     *
     * @return destinationId as UUID
     */
    public UUID getDestinationId() {
        return submission.getDestinationId();
    }

    /**
     * Get the submissions submissionId.
     *
     * @return submissionId as UUID
     */
    public UUID getSubmissionId() {
        return submission.getSubmissionId();
    }
}
