package dev.fitko.fitconnect.client;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.config.BuildInfo;
import dev.fitko.fitconnect.api.config.Environment;
import dev.fitko.fitconnect.api.config.EnvironmentName;
import dev.fitko.fitconnect.api.config.SenderConfig;
import dev.fitko.fitconnect.api.config.SubscriberConfig;
import dev.fitko.fitconnect.api.domain.auth.OAuthToken;
import dev.fitko.fitconnect.api.domain.model.event.Event;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.EventStatus;
import dev.fitko.fitconnect.api.domain.model.event.problems.submission.InvalidEventLog;
import dev.fitko.fitconnect.api.domain.model.metadata.ContentStructure;
import dev.fitko.fitconnect.api.domain.model.metadata.Hash;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.PublicServiceType;
import dev.fitko.fitconnect.api.domain.model.metadata.SignatureType;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Purpose;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.route.Area;
import dev.fitko.fitconnect.api.domain.model.route.Route;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.client.factory.ClientFactory;
import dev.fitko.fitconnect.client.router.DestinationSearch;
import dev.fitko.fitconnect.client.sender.EncryptedSubmissionBuilder;
import dev.fitko.fitconnect.client.sender.SubmissionBuilder;
import dev.fitko.fitconnect.client.sender.model.AttachmentPayload;
import dev.fitko.fitconnect.client.subscriber.ReceivedSubmission;
import dev.fitko.fitconnect.core.auth.DefaultOAuthService;
import dev.fitko.fitconnect.core.crypto.HashService;
import dev.fitko.fitconnect.core.crypto.JWECryptoService;
import dev.fitko.fitconnect.core.http.RestService;
import org.apache.tika.mime.MimeTypes;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.condition.EnabledIfEnvironmentVariable;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.text.ParseException;
import java.time.Duration;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;
import java.util.stream.Collectors;

import static dev.fitko.fitconnect.api.config.SchemaConfig.METADATA_V_1_0_0;
import static org.awaitility.Awaitility.await;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.empty;
import static org.hamcrest.Matchers.hasItems;
import static org.hamcrest.Matchers.hasSize;
import static org.hamcrest.Matchers.is;
import static org.hamcrest.Matchers.lessThanOrEqualTo;
import static org.hamcrest.Matchers.not;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;

/**
 * This test uses real credentials and endpoints stored in gitlab ci variables to test the sdk against a real system
 */
@EnabledIfEnvironmentVariable(named="TEST_DESTINATION_ID", matches = ".*")
@EnabledIfEnvironmentVariable(named="SENDER_CLIENT_ID", matches = ".*")
@EnabledIfEnvironmentVariable(named="SENDER_CLIENT_SECRET", matches = ".*")
@EnabledIfEnvironmentVariable(named="SUBSCRIBER_CLIENT_ID", matches = ".*")
@EnabledIfEnvironmentVariable(named="SUBSCRIBER_CLIENT_SECRET", matches = ".*")
class ClientIntegrationTest {

    @Nested
    class SendSubmissionTests {

        @Test
        void testSendAndConfirmCycle() {

            // Given
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab("TESTING", true);

            final var submission = SubmissionBuilder.Builder()
                    .withAttachment(new File("src/test/resources/attachment.txt"))
                    .withJsonData("{ \"data\": \"Beispiel Fachdaten\" }")
                    .withDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                    .withServiceType("Test Service", "urn:de:fim:leika:leistung:99400048079000")
                    .build();

            final var sentSubmission = ClientFactory.senderClient(config).submit(submission);

            assertNotNull(sentSubmission);

            // When
            final ReceivedSubmission receivedSubmission =
                    ClientFactory.subscriberClient(config)
                            .requestSubmission(sentSubmission.getSubmissionId());

            // Then
            assertNotNull(receivedSubmission);
            assertThat(receivedSubmission.getData(), is("{ \"data\": \"Beispiel Fachdaten\" }"));
            assertThat(receivedSubmission.getDataSchemaUri(), is(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json")));
            assertThat(receivedSubmission.getDataMimeType().toString(), is("application/json"));
            assertThat(new String(receivedSubmission.getAttachments().get(0).getData()), is("Test attachment"));
        }

        @Test
        void testSendAndConfirmCycleWithEncryptedData() throws ParseException, IOException {

            // Given
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab("TESTING", true);

            final UUID destinationId = UUID.fromString(System.getenv("TEST_DESTINATION_ID"));

            final CryptoService cryptoService = new JWECryptoService(new HashService());
            final SenderClient client = ClientFactory.senderClient(config);

            final Optional<String> publicKey = client.getPublicKey(destinationId);
            assertTrue(publicKey.isPresent());

            final RSAKey encryptionKey = RSAKey.parse(publicKey.get());

            final String jsonData = "{ \"data\": \"Beispiel Fachdaten\" }";
            final File attachmentFile = new File("src/test/resources/attachment.txt");
            final String attachmentData = Files.readString(attachmentFile.toPath());

            final String encryptedData = cryptoService.encryptString(encryptionKey, jsonData);
            final String encryptedAttachment = cryptoService.encryptString(encryptionKey, attachmentData);

            final var submissionSchema = new SubmissionSchema();
            submissionSchema.setSchemaUri(URI.create("https://schema.fitko.de/fim/s00000000009_1.0.0.schema.json"));
            submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

            final var dataHash = new Hash();
            dataHash.setSignatureType(SignatureType.SHA_512);
            dataHash.setContent(cryptoService.hashBytes(jsonData.getBytes(StandardCharsets.UTF_8)));

            final var data = new Data();
            data.setHash(dataHash);
            data.setSubmissionSchema(submissionSchema);

            final var attachmentPayload = AttachmentPayload.builder()
                    .encryptedData(encryptedAttachment)
                    .file(attachmentFile)
                    .mimeType(MimeTypes.PLAIN_TEXT)
                    .attachmentId(UUID.randomUUID())
                    .hashedData(cryptoService.hashBytes(attachmentData.getBytes(StandardCharsets.UTF_8)))
                    .build();

            final var publicServiceType = new PublicServiceType();
            publicServiceType.setName("Test Service");
            publicServiceType.setIdentifier("urn:de:fim:leika:leistung:99400048079000");

            final var attachment = new Attachment();
            attachment.setAttachmentId(attachmentPayload.getAttachmentId());
            attachment.setPurpose(Purpose.ATTACHMENT);
            attachment.setFilename(attachmentPayload.getFile().getName());
            attachment.setMimeType(attachmentPayload.getMimeType());

            final var attachmentHash = new Hash();
            attachmentHash.setContent(attachmentPayload.getHashedData());
            attachmentHash.setSignatureType(SignatureType.SHA_512);
            attachment.setHash(attachmentHash);

            final var contentStructure = new ContentStructure();
            contentStructure.setAttachments(List.of(attachment));
            contentStructure.setData(data);

            final var metadata = new Metadata();
            metadata.setSchema(METADATA_V_1_0_0.toString());
            metadata.setContentStructure(contentStructure);
            metadata.setPublicServiceType(publicServiceType);

            final String encryptedMetadata = cryptoService.encryptBytes(encryptionKey, new ObjectMapper().writeValueAsBytes(metadata));

            // When
            final var submission = EncryptedSubmissionBuilder.Builder()
                    .withEncryptedAttachment(attachmentPayload.getAttachmentId(), attachmentPayload.getEncryptedData())
                    .withEncryptedData(encryptedData)
                    .withEncryptedMetadata(encryptedMetadata)
                    .withDestination(destinationId)
                    .withServiceType("Test Service", "urn:de:fim:leika:leistung:99400048079000")
                    .build();

            final var sentSubmission = ClientFactory.senderClient(config).submit(submission);

            assertNotNull(sentSubmission);

            final ReceivedSubmission receivedSubmission = ClientFactory.subscriberClient(config).requestSubmission(sentSubmission.getSubmissionId());

            // Then
            assertNotNull(receivedSubmission);
            assertThat(receivedSubmission.getData(), is(jsonData));
            assertThat(receivedSubmission.getDataMimeType().toString(), is("application/json"));
            assertThat(new String(receivedSubmission.getAttachments().get(0).getData()), is("Test attachment"));
        }

        @Test
        void testAbortedSendSubmissionWithKeyValidationNotSilent() {

            // Given
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab("PROD", false);

            // When
            final var submission = SubmissionBuilder.Builder()
                    .withJsonData("{ \"data\": \"Beispiel Fachdaten\" }")
                    .withDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                    .withServiceType("Test Service", "urn:de:fim:leika:leistung:99400048079000")
                    .build();

            final var sentSubmission = ClientFactory.senderClient(config).submit(submission);

            // Then
            assertNull(sentSubmission);
        }
    }

    @Nested
    class ReceiveSubmissionTests {

        @Test
        void testListSubmissions() {

            // Given
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab("TESTING", true);

            final var senderClient = ClientFactory.senderClient(config);
            final var subscriberClient = ClientFactory.subscriberClient(config);

            final UUID destinationId = UUID.fromString(System.getenv("TEST_DESTINATION_ID"));
            final String leikaKey = "urn:de:fim:leika:leistung:99400048079000";
            final String serviceName = "Test Service";

            final var submissionOne = SubmissionBuilder.Builder()
                    .withJsonData("{ \"data\": \"Beispiel Fachdaten 1\" }")
                    .withDestination(destinationId)
                    .withServiceType(serviceName, leikaKey)
                    .build();

            final var submissionTwo = SubmissionBuilder.Builder()
                    .withJsonData("{ \"data\": \"Beispiel Fachdaten 2\" }")
                    .withDestination(destinationId)
                    .withServiceType(serviceName, leikaKey)
                    .build();

            final var sentSubmissionOne = senderClient.submit(submissionOne);
            final var sentSubmissionTwo = senderClient.submit(submissionTwo);

            assertNotNull(sentSubmissionOne);
            assertNotNull(sentSubmissionTwo);

            // When
            final Set<SubmissionForPickup> submissions = subscriberClient.getAvailableSubmissions(destinationId);

            // Then
            assertFalse(submissions.isEmpty());

            final List<UUID> submissionIds = submissions.stream().map(SubmissionForPickup::getSubmissionId).collect(Collectors.toList());

            assertThat(submissionIds, hasItems(sentSubmissionOne.getSubmissionId(), sentSubmissionTwo.getSubmissionId()));

            // remove by confirming
            subscriberClient.requestSubmission(sentSubmissionOne.getSubmissionId()).acceptSubmission();
            subscriberClient.requestSubmission(sentSubmissionTwo.getSubmissionId()).acceptSubmission();
        }
    }

    @Nested
    class EventTests {

        @Test
        void testRejectEvent() {

            // Given
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab("TESTING", true);

            final var submission = SubmissionBuilder.Builder()
                    .withAttachment(new File("src/test/resources/attachment.txt"))
                    .withJsonData("{ \"data\": \"Beispiel Fachdaten\" }")
                    .withDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                    .withServiceType("Test Service", "urn:de:fim:leika:leistung:99400048079000")
                    .build();

            final var sentSubmission = ClientFactory.senderClient(config).submit(submission);

            assertNotNull(sentSubmission);

            // When
            final var subscriberClient = ClientFactory.subscriberClient(config);

            final var sentSubmissionId = sentSubmission.getSubmissionId();

            // reject and remove
            subscriberClient.requestSubmission(sentSubmissionId).rejectSubmission(List.of(new InvalidEventLog()));

            // second attempt to receive and reject the submission should return an empty result
            // since the submission is gone after being rejected
            assertNull(subscriberClient.requestSubmission(sentSubmissionId));
        }

        @Test
        void testAcceptEvent() {

            // Given
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab("TESTING", true);

            final var submission = SubmissionBuilder.Builder()
                    .withAttachment(new File("src/test/resources/attachment.txt"))
                    .withJsonData("{ \"data\": \"Beispiel Fachdaten\" }")
                    .withDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                    .withServiceType("Test Service", "urn:de:fim:leika:leistung:99400048079000")
                    .build();

            final var sentSubmission = ClientFactory.senderClient(config).submit(submission);

            assertNotNull(sentSubmission);

            // When
            final var subscriberClient = ClientFactory.subscriberClient(config);

            final var sentSubmissionId = sentSubmission.getSubmissionId();

            // accept and remove
            subscriberClient.requestSubmission(sentSubmissionId).acceptSubmission();

            // second attempt to receive the submission should return an empty result
            // since the submission is gone after being accepted
            assertNull(subscriberClient.requestSubmission(sentSubmissionId));
        }

        @Test
        void testReadEventLogFromSender() {

            // Given
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab("TESTING", true);

            final var submission = SubmissionBuilder.Builder()
                    .withJsonData("{ \"data\": \"Beispiel Fachdaten\" }")
                    .withDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                    .withServiceType("Test Service", "urn:de:fim:leika:leistung:99400048079000")
                    .build();

            final var sentSubmission = ClientFactory.senderClient(config).submit(submission);

            assertNotNull(sentSubmission);

            // When
            await().atMost(Duration.ofSeconds(30)).until(() ->
            {
                final UUID destinationId = sentSubmission.getDestinationId();
                final UUID caseId = sentSubmission.getCaseId();

                final List<EventLogEntry> senderEventLog = ClientFactory.senderClient(config).getEventLog(caseId, destinationId);
                final List<Event> senderEvents = senderEventLog.stream().map(EventLogEntry::getEvent).collect(Collectors.toList());

                // Then
                return senderEvents.equals(List.of(Event.CREATE, Event.SUBMIT));
            });
        }

        @Test
        void testReadEventLogFromSubscriber() {

            // Given
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab("TESTING", true);

            final var submission = SubmissionBuilder.Builder()
                    .withAttachment(new File("src/test/resources/attachment.txt"))
                    .withJsonData("{ \"data\": \"Beispiel Fachdaten\" }")
                    .withDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                    .withServiceType("Test Service", "urn:de:fim:leika:leistung:99400048079000")
                    .build();

            final var sentSubmission = ClientFactory.senderClient(config).submit(submission);

            assertNotNull(sentSubmission);

            final var receivedSubmission = ClientFactory.subscriberClient(config)
                    .requestSubmission(sentSubmission.getSubmissionId());

            assertNotNull(receivedSubmission);

            // When
            await().atMost(Duration.ofSeconds(30)).until(() ->
            {
                final UUID destinationId = receivedSubmission.getDestinationId();
                final UUID caseId = receivedSubmission.getCaseId();

                final List<EventLogEntry> subscriberEventLog = ClientFactory.subscriberClient(config).getEventLog(caseId, destinationId);
                final List<Event> subscriberEvents = subscriberEventLog.stream().map(EventLogEntry::getEvent).collect(Collectors.toList());

                // Then
                return subscriberEvents.equals(List.of(Event.CREATE, Event.SUBMIT));
            });
        }

        @Test
        void testReadSubmissionStatusWithAuthTagEventValidationFromSender() {

            // Given
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab("TESTING", true);

            final SenderClient senderClient = ClientFactory.senderClient(config);

            final var submission = SubmissionBuilder.Builder()
                    .withAttachment(new File("src/test/resources/attachment.txt"))
                    .withJsonData("{ \"data\": \"Beispiel Fachdaten\" }")
                    .withDestination(UUID.fromString(System.getenv("TEST_DESTINATION_ID")))
                    .withServiceType("Test Service", "urn:de:fim:leika:leistung:99400048079000")
                    .build();

            final var sentSubmission = ClientFactory.senderClient(config).submit(submission);

            assertNotNull(sentSubmission);

            await().atMost(Duration.ofSeconds(30)).until(() ->
            {
                // When
                final EventStatus statusForSubmission = senderClient.getStatusForSubmission(sentSubmission);

                // Then
                return statusForSubmission.getStatus().equals(Event.SUBMIT.getState());
            });
        }

    }

    @Nested
    class RoutingTests {

        @Test
        void testFindDestinationsWithRegionalKey() {

            // Given
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab("TESTING", true);

            final RoutingClient routingClient = ClientFactory.routingClient(config);

            final DestinationSearch search = DestinationSearch.Builder()
                    .withLeikaKey("99123456760610")
                    .withArs("064350014014")
                    .build();

            // When
            final List<Route> routes = routingClient.findDestinations(search);

            // Then
            assertThat(routes, hasSize(1));
            assertThat(routes.get(0).getDestinationId(), is(UUID.fromString("d40e7b13-da98-4b09-9e16-bbd61ca81510")));
        }

        @Test
        void testFindDestinationsWithAreaId() {

            // Given
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab("TESTING", true);

            final RoutingClient routingClient = ClientFactory.routingClient(config);

            final DestinationSearch search = DestinationSearch.Builder()
                    .withLeikaKey("99123456760610")
                    .withAreaId("931")
                    .build();

            // When
            final List<Route> routes = routingClient.findDestinations(search);

            // Then
            assertThat(routes, hasSize(1));
            assertThat(routes.get(0).getDestinationId(), is(UUID.fromString("d40e7b13-da98-4b09-9e16-bbd61ca81510")));
        }

        @Test
        void testFindDestinationsWithMultipleAreaSearchCriteria() {

            // Given
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab("TESTING", true);

            final RoutingClient routingClient = ClientFactory.routingClient(config);

            final DestinationSearch search = DestinationSearch.Builder()
                    .withLeikaKey("99123456760610")
                    .withArs("064350014014")
                    .withAreaId("1234")
                    .build();

            // When
            final RestApiException exception = assertThrows(RestApiException.class, () -> routingClient.findDestinations(search));

            // Then
            assertThat(exception.getMessage(), containsString("Only one of ars, ags or areaId must be specified"));
        }

        @Test
        void testFindAreaWithMultipleSearchCriteria() {

            // Given
            final ApplicationConfig config = getConfigWithCredentialsFromGitlab("TESTING", true);

            final RoutingClient routingClient = ClientFactory.routingClient(config);

            // When
            final List<Area> areas = routingClient.findAreas(List.of("Leip*", "04229"), 0, 10);

            // Then
            assertThat(areas, is(not(empty())));
            assertThat(areas.size(), is(lessThanOrEqualTo(10)));
            assertTrue(areas.stream().anyMatch(area -> area.getName().equals("Leipzig")));
        }

    }

    @Nested
    class AuthenticationTests {
        @Test
        void retrieveAuthenticationToken() {

            // Given
            final var tokenUrl = "https://auth-testing.fit-connect.fitko.dev/token";
            final var clientId = System.getenv("SUBSCRIBER_CLIENT_ID");
            final var secret = System.getenv("SUBSCRIBER_CLIENT_SECRET");

            final RestService restService = new RestService(new BuildInfo());

            final var authService = new DefaultOAuthService(restService.getRestTemplate(), clientId, secret, tokenUrl);

            // When
            final OAuthToken token = authService.getCurrentToken();

            // Then
            assertNotNull(token);
            assertNull(token.getError());
            assertNotNull(token.getAccessToken());
            assertEquals(1800, token.getExpiresIn());
        }
    }


    private ApplicationConfig getConfigWithCredentialsFromGitlab(final String environmentName, final boolean allowInsecurePublicKey) {

        final var sender = new SenderConfig(System.getenv("SENDER_CLIENT_ID"), System.getenv("SENDER_CLIENT_SECRET"));

        final var subscriber = SubscriberConfig.builder()
                .clientId(System.getenv("SUBSCRIBER_CLIENT_ID"))
                .clientSecret(System.getenv("SUBSCRIBER_CLIENT_SECRET"))
                .privateDecryptionKeyPath("src/test/resources/private_decryption_test_key.json")
                .privateSigningKeyPath("src/test/resources/private_test_signing_key.json")
                .build();

        final var envName = new EnvironmentName(environmentName);
        final Environment env = getEnvironment(allowInsecurePublicKey);

        return ApplicationConfig.builder()
                .senderConfig(sender)
                .subscriberConfig(subscriber)
                .environments(Map.of(envName, env))
                .activeEnvironment(envName)
                .build();
    }

    private static Environment getEnvironment(final boolean allowInsecurePublicKey) {
        final String authBaseUrl = "https://auth-testing.fit-connect.fitko.dev";
        final String routingBaseUrl = "https://routing-api-testing.fit-connect.fitko.dev";
        final String selfServicePortalUrl = "https://portal.auth-testing.fit-connect.fitko.dev";
        final String submissionBaseUrl = "https://submission-api-testing.fit-connect.fitko.dev";
        return new Environment(authBaseUrl, routingBaseUrl, submissionBaseUrl, selfServicePortalUrl, allowInsecurePublicKey);
    }
}
