package dev.fitko.fitconnect.client.factory;

import dev.fitko.fitconnect.api.config.*;
import dev.fitko.fitconnect.api.exceptions.InitializationException;
import dev.fitko.fitconnect.api.exceptions.InvalidKeyException;
import org.junit.jupiter.api.Test;

import java.util.Map;

import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;

class ClientFactoryTest {

    @Test
    void testMissingConfiguration() {
        assertThrows(InitializationException.class, ClientFactory::senderClient);
        assertThrows(InitializationException.class, ClientFactory::subscriberClient);
        assertThrows(InitializationException.class, ClientFactory::routingClient);
    }

    @Test
    void testSenderClientConstruction() {

        final var envName = new EnvironmentName("DEV");
        final var environments = Map.of(envName, new Environment("https://auth", "", "", "", true));

        final var sender = new SenderConfig("123", "abc");

        final var senderConfig = ApplicationConfig.builder()
                .httpProxyHost("https://proxy.fitco.de")
                .httpProxyPort(0)
                .environments(environments)
                .senderConfig(sender)
                .activeEnvironment(envName)
                .build();

        assertNotNull(ClientFactory.senderClient(senderConfig));
    }

    @Test
    void testSubscriberClientConstruction() {

        final var envName = new EnvironmentName("DEV");
        final var environments = Map.of(envName, new Environment("https://auth", "", "", "", true));

        final var subscriber = SubscriberConfig.builder()
                .clientSecret("123")
                .clientSecret("abc")
                .privateDecryptionKeyPath("src/test/resources/private_decryption_test_key.json")
                .privateSigningKeyPath("src/test/resources/private_test_signing_key.json")
                .build();

        final var subscriberConfig = ApplicationConfig.builder()
                .httpProxyHost("https://proxy.fitco.de")
                .httpProxyPort(0)
                .environments(environments)
                .subscriberConfig(subscriber)
                .activeEnvironment(envName)
                .build();

        assertNotNull(ClientFactory.subscriberClient(subscriberConfig));
    }

    @Test
    void testRoutingClientConstruction() {

        final var envName = new EnvironmentName("DEV");
        final var environment = new Environment("", "https://routing.fitko.fitconnect.de", "", "", true);

        final var senderConfig = new SenderConfig("1234", "abcd");

        final var routingConfig = ApplicationConfig.builder()
                .environments(Map.of(envName, environment))
                .activeEnvironment(envName)
                .senderConfig(senderConfig)
                .build();

        assertNotNull(ClientFactory.routingClient(routingConfig));
    }

    @Test
    void testSigningKeyCannotBeParsed() {

        final var envName = new EnvironmentName("DEV");
        final var environments = Map.of(envName, new Environment("https://auth", "", "", "", true));

        final var subscriberConfig = SubscriberConfig.builder()
                .clientSecret("123")
                .clientSecret("abc")
                .privateDecryptionKeyPath("src/test/resources/private_decryption_test_key.json")
                .privateSigningKeyPath("src/test/resources/invalid_signing_key.json")
                .build();

        final var config = ApplicationConfig.builder()
                .httpProxyHost("https://proxy.fitco.de")
                .httpProxyPort(0)
                .subscriberConfig(subscriberConfig)
                .environments(environments)
                .activeEnvironment(envName)
                .build();

        assertThrows(
                InvalidKeyException.class,
                () -> ClientFactory.subscriberClient(config),
                "Expected new ClientFactoryTest() to throw, but nothing was thrown"
        );
    }

    @Test
    void testDecryptionKeyCannotBeParsed() {

        final var envName = new EnvironmentName("DEV");
        final var environments = Map.of(envName, new Environment("https://auth", "", "", "", true));

        final var subscriberConfigWithoutKey = SubscriberConfig.builder()
                .clientSecret("123")
                .clientSecret("abc")
                .privateDecryptionKeyPath("src/test/resources/invalid_private_key.json")
                .privateSigningKeyPath("src/test/resources/private_test_signing_key.json")
                .build();

        final var config = ApplicationConfig.builder()
                .httpProxyHost("https://proxy.fitco.de")
                .httpProxyPort(0)
                .subscriberConfig(subscriberConfigWithoutKey)
                .environments(environments)
                .activeEnvironment(envName)
                .build();

        assertThrows(
                InvalidKeyException.class,
                () -> ClientFactory.subscriberClient(config),
                "Expected new ClientFactoryTest() to throw, but nothing was thrown"
        );
    }
}