package dev.fitko.fitconnect.core;

import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.EventStatus;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.submission.CreateSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.api.services.events.EventLogService;
import dev.fitko.fitconnect.api.services.keys.KeyService;
import dev.fitko.fitconnect.api.services.submission.SubmissionService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.UUID;

public class SubmissionSender implements Sender {

    private static final Logger LOGGER = LoggerFactory.getLogger(SubmissionSender.class);

    private final ValidationService validationService;
    private final CryptoService cryptoService;
    private final SubmissionService submissionService;
    private final EventLogService eventLogService;

    private final KeyService keyService;

    public SubmissionSender(final SubmissionService submissionService,
                            final EventLogService eventLogService,
                            final CryptoService cryptoService,
                            final ValidationService validationService,
                            final KeyService keyService) {

        this.submissionService = submissionService;
        this.eventLogService = eventLogService;
        this.cryptoService = cryptoService;
        this.validationService = validationService;
        this.keyService = keyService;
    }

    @Override
    public ValidationResult validatePublicKey(final RSAKey publicKey) {
        LOGGER.info("Validating public key integrity");
        return validationService.validateEncryptionPublicKey(publicKey);
    }

    @Override
    public ValidationResult validateMetadata(final Metadata metadata) {
        LOGGER.info("Validating metadata");
        return validationService.validateMetadataSchema(metadata);
    }

    @Override
    public ValidationResult validateCallback(final String hmac, final Long timestamp, final String httpBody, final String callbackSecret) {
        LOGGER.info("Validating callback integrity");
        return validationService.validateCallback(hmac, timestamp, httpBody, callbackSecret);
    }

    @Override
    public String encryptBytes(final RSAKey publicKey, final byte[] data) {
        LOGGER.info("Encrypting {} bytes", data.length);
        return cryptoService.encryptBytes(publicKey, data);
    }

    @Override
    public String encryptObject(final RSAKey encryptionKey, final Object obj) {
        LOGGER.info("Encrypting object");
        return cryptoService.encryptObject(encryptionKey, obj);
    }


    @Override
    public String createHash(final byte[] data) {
        LOGGER.info("Creating hash of {} bytes", data.length);
        return cryptoService.hashBytes(data);
    }

    @Override
    public SubmissionForPickup createSubmission(final CreateSubmission submission) {
        LOGGER.info("Announcing new submission for destination {}", submission.getDestinationId());
        return submissionService.announceSubmission(submission);
    }

    @Override
    public void uploadAttachment(final UUID submissionId, final UUID attachmentId, final String encryptedAttachment) {
        LOGGER.info("Uploading attachment {} for submission {}", attachmentId, submissionId);
        submissionService.uploadAttachment(submissionId, attachmentId, encryptedAttachment);
    }

    @Override
    public Destination getDestination(final UUID destinationId) {
        LOGGER.info("Loading  destination {}", destinationId);
        return submissionService.getDestination(destinationId);
    }

    @Override
    public RSAKey getEncryptionKeyForDestination(final UUID destinationId) {
        LOGGER.info("Loading encryption key for destination id {}", destinationId);
        return keyService.getPublicEncryptionKey(destinationId);
    }


    @Override
    public Submission sendSubmission(final SubmitSubmission submission) {
        LOGGER.info("Sending submission {}", submission.getSubmissionId());
        return submissionService.sendSubmission(submission);
    }

    @Override
    public ValidationResult validateJsonFormat(final String json) {
        LOGGER.info("Validating data json format");
        return validationService.validateJsonFormat(json);
    }

    @Override
    public ValidationResult validateXmlFormat(final String xml) {
        LOGGER.info("Validating data xml format");
        return validationService.validateXmlFormat(xml);
    }

    @Override
    public List<EventLogEntry> getEventLog(final UUID caseId, final UUID destinationId) {
        LOGGER.info("Loading event log for destination {}", destinationId);
        return eventLogService.getEventLog(caseId, destinationId);
    }

    @Override
    public EventStatus getLastedEvent(final UUID destinationId, final UUID caseId, final UUID submissionId, final AuthenticationTags authenticationTags) {
        LOGGER.info("Loading latest status for submission {}", submissionId);
        return eventLogService.getLastedEvent(destinationId, caseId, submissionId, authenticationTags);
    }
}
