package dev.fitko.fitconnect.core.crypto;

import dev.fitko.fitconnect.api.exceptions.EncryptionException;
import dev.fitko.fitconnect.api.exceptions.InitializationException;
import dev.fitko.fitconnect.api.services.crypto.MessageDigestService;
import dev.fitko.fitconnect.core.util.StopWatch;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.crypto.Mac;
import javax.crypto.spec.SecretKeySpec;
import java.security.InvalidKeyException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;

import static dev.fitko.fitconnect.core.crypto.CryptoConstants.DEFAULT_HASH_ALGORITHM;
import static dev.fitko.fitconnect.core.crypto.CryptoConstants.DEFAULT_HMAC_ALGORITHM;

public class HashService implements MessageDigestService {

    private static final Logger LOGGER = LoggerFactory.getLogger(HashService.class);

    private final MessageDigest messageDigest;

    public HashService() {
        try {
            this.messageDigest = MessageDigest.getInstance(DEFAULT_HASH_ALGORITHM);
        } catch (final NoSuchAlgorithmException e) {
            throw new InitializationException(e.getMessage(), e);
        }
    }

    HashService(final MessageDigest messageDigest) {
        this.messageDigest = messageDigest;
    }

    @Override
    public byte[] createHash(final byte[] data) {
        this.messageDigest.reset();
        final var start = StopWatch.start();
        final byte[] hash = this.messageDigest.digest(data);
        LOGGER.info("Hashing {} bytes took {}", data.length, StopWatch.stopWithFormattedTime(start));
        return hash;

    }

    @Override
    public boolean verify(final byte[] originalHash, final byte[] data) {
        final byte[] newHash = createHash(data);
        return compareHashes(originalHash, newHash);
    }

    @Override
    public String toHexString(final byte[] hash) {
        final StringBuilder hexStringBuffer = new StringBuilder();
        for (int i = 0; i < hash.length; i++) {
            hexStringBuffer.append(byteToHex(hash[i]));
        }
        return hexStringBuffer.toString();
    }

    @Override
    public byte[] fromHexString(final String hexString) {
        if (hexString.length() % 2 == 1) {
            throw new IllegalArgumentException("Invalid hexadecimal String.");
        }

        final byte[] bytes = new byte[hexString.length() / 2];
        for (int i = 0; i < hexString.length(); i += 2) {
            bytes[i / 2] = hexToByte(hexString.substring(i, i + 2));
        }
        return bytes;
    }

    @Override
    public String calculateHMAC(String data, String key) {

        try {
            String hmacAlgorithm = DEFAULT_HMAC_ALGORITHM;
            SecretKeySpec secretKeySpec = new SecretKeySpec(key.getBytes(), hmacAlgorithm);
            Mac mac = Mac.getInstance(hmacAlgorithm);
            mac.init(secretKeySpec);
            return toHexString(mac.doFinal(data.getBytes()));
        } catch (NoSuchAlgorithmException | InvalidKeyException e) {
            throw new EncryptionException("Calculation of HMAC failed.", e);
        }
    }

    private byte hexToByte(final String hexString) {
        final int firstDigit = toDigit(hexString.charAt(0));
        final int secondDigit = toDigit(hexString.charAt(1));
        return (byte) ((firstDigit << 4) + secondDigit);
    }

    private int toDigit(final char hexChar) {
        final int digit = Character.digit(hexChar, 16);
        if (digit == -1) {
            throw new IllegalArgumentException("Invalid Hexadecimal Character: " + hexChar);
        }
        return digit;
    }

    private String byteToHex(final byte num) {
        final char[] hexDigits = new char[2];
        hexDigits[0] = Character.forDigit((num >> 4) & 0xF, 16);
        hexDigits[1] = Character.forDigit((num & 0xF), 16);
        return new String(hexDigits);
    }

    private boolean compareHashes(final byte[] originalHash, final byte[] comparisonHash) {
        int diff = originalHash.length ^ comparisonHash.length;
        for (int i = 0; i < originalHash.length && i < comparisonHash.length; i++) {
            diff |= originalHash[i] ^ comparisonHash[i];
        }
        return diff == 0;
    }

}
