package dev.fitko.fitconnect.core.events;

import com.nimbusds.jwt.SignedJWT;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.domain.model.event.Event;
import dev.fitko.fitconnect.api.domain.model.event.EventLog;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.EventStatus;
import dev.fitko.fitconnect.api.domain.model.event.SubmissionState;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.validation.ValidationContext;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.EventLogException;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import dev.fitko.fitconnect.api.services.auth.OAuthService;
import dev.fitko.fitconnect.api.services.events.EventLogService;
import dev.fitko.fitconnect.api.services.events.EventLogVerificationService;
import dev.fitko.fitconnect.core.util.EventLogUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.web.client.RestClientException;
import org.springframework.web.client.RestTemplate;

import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.UUID;
import java.util.function.Function;
import java.util.stream.Collectors;

public class EventLogApiService implements EventLogService {

    private static final Logger LOGGER = LoggerFactory.getLogger(EventLogApiService.class);

    private final ApplicationConfig config;
    private final OAuthService authService;
    private final RestTemplate restTemplate;
    private final EventLogVerificationService eventLogVerifier;

    public EventLogApiService(final ApplicationConfig config,
                              final OAuthService authService,
                              final RestTemplate restTemplate,
                              final EventLogVerificationService eventLogVerifier) {
        this.config = config;
        this.authService = authService;
        this.restTemplate = restTemplate;
        this.eventLogVerifier = eventLogVerifier;
    }

    @Override
    public List<EventLogEntry> getEventLog(final UUID caseId, final UUID destinationId) {
        final EventLog eventLog = loadEventLog(caseId);
        final List<SignedJWT> events = EventLogUtil.getJWTSFromEvents(eventLog.getEventLogs());
        final ValidationContext ctx = new ValidationContext(destinationId, caseId);
        validateLog(eventLogVerifier.validateEventLogs(ctx, events));
        return EventLogUtil.mapEventLogToEntries(eventLog);
    }

    @Override
    public EventStatus getLastedEvent(final UUID destinationId, final UUID caseId, final UUID submissionId, final AuthenticationTags authenticationTags) throws RestApiException, EventLogException {
        final EventLog eventLog = loadEventLog(caseId);
        final SignedJWT latestEvent = getLatestEventForSubmission(submissionId, eventLog);
        if (latestEvent == null) {
            LOGGER.info("No events found for submission {}", submissionId);
            return new EventStatus();
        }
        final var ctx = new ValidationContext(destinationId, caseId, authenticationTags);
        validateLog(eventLogVerifier.validateEventLogs(ctx, List.of(latestEvent)));
        final EventLogEntry eventLogEntry = EventLogUtil.eventToEntry(latestEvent);
        return new EventStatus(eventLogEntry.getEvent().getState(), eventLogEntry.getProblems());
    }

    @Override
    public void sendEvent(final UUID caseId, final String signedAndSerializedSET) {
        final String url = config.getEventsEndpoint();
        final HttpHeaders headers = getHttpHeaders("application/jose");
        final HttpEntity<String> entity = new HttpEntity<>(signedAndSerializedSET, headers);
        try {
            restTemplate.exchange(url, HttpMethod.POST, entity, Void.class, caseId);
        } catch (final RestClientException e) {
            throw new RestApiException("Sending event failed", e);
        }
    }

    private EventLog loadEventLog(final UUID caseId) {
        final String url = config.getEventsEndpoint();
        final HttpHeaders headers = getHttpHeaders();
        final HttpEntity<String> entity = new HttpEntity<>(headers);
        try {
            return restTemplate.exchange(url, HttpMethod.GET, entity, EventLog.class, caseId).getBody();
        } catch (final RestClientException e) {
            throw new RestApiException("EventLog query failed", e);
        }
    }

    private HttpHeaders getHttpHeaders(final String contentType) {
        return getHttpHeaders(MediaType.parseMediaType(contentType));
    }

    private HttpHeaders getHttpHeaders() {
        return getHttpHeaders(MediaType.APPLICATION_JSON);
    }

    private HttpHeaders getHttpHeaders(final MediaType mediaType) {
        final var headers = new HttpHeaders();
        headers.setBearerAuth(authService.getCurrentToken().getAccessToken());
        headers.setContentType(mediaType);
        headers.setAcceptCharset(List.of(StandardCharsets.UTF_8));
        return headers;
    }

    private SignedJWT getLatestEventForSubmission(final UUID submissionId, final EventLog eventLog) {

        final List<SignedJWT> signedJWTS = EventLogUtil.getJWTSFromEvents(eventLog.getEventLogs());
        final SortedMap<Event, SignedJWT> eventsOrderedByState = getOrderedEvents(submissionId, signedJWTS);
        final SignedJWT latestEvent = getLatestEvent(submissionId, signedJWTS);

        return latestEvent != null ? getLatestEventIfInAllowedOrder(eventsOrderedByState, latestEvent) : null;
    }

    private SignedJWT getLatestEventIfInAllowedOrder(final SortedMap<Event, SignedJWT> eventsOrderedByState, final SignedJWT latestEvent) {
        final SubmissionState latestAllowedState = eventsOrderedByState.lastKey().getState();
        final SubmissionState currentLatestState = EventLogUtil.getEventFromJWT(latestEvent).getState();
        if (latestAllowedState != currentLatestState) {
            final String allowedOrder = eventsOrderedByState.keySet().stream().map(e -> e.getState().getName().toUpperCase()).collect(Collectors.joining("|"));
            throw new EventLogException("Allowed transition [" + allowedOrder + "] should have [" + latestAllowedState + "] as latest state but was [" + currentLatestState + "]");
        }
        return latestEvent;
    }

    private SignedJWT getLatestEvent(final UUID submissionId, final List<SignedJWT> signedJWTS) {
        return signedJWTS.stream()
                .filter(EventLogUtil.eventEqualsSubmissionId(submissionId))
                .reduce((first, second) -> second)
                .orElse(null);
    }

    private SortedMap<Event, SignedJWT> getOrderedEvents(final UUID submissionId, final List<SignedJWT> signedJWTS) {
        return new TreeMap<>(signedJWTS.stream()
                .filter(EventLogUtil.eventEqualsSubmissionId(submissionId))
                .collect(Collectors.toMap(EventLogUtil::getEventFromJWT, Function.identity(), (e1, e2) -> e1)));
    }

    private void validateLog(final List<ValidationResult> results) {
        if (!results.isEmpty()) {
            final String errorMessages = results.stream()
                    .map(ValidationResult::getError)
                    .map(Throwable::getMessage)
                    .distinct()
                    .collect(Collectors.joining("\n"));
            throw new EventLogException("Event log contains invalid entries: \n" + errorMessages);
        }
    }
}
