package dev.fitko.fitconnect.core.keys;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.jwk.ApiJwk;
import dev.fitko.fitconnect.api.domain.model.jwk.ApiJwkSet;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.InvalidKeyException;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import dev.fitko.fitconnect.api.services.auth.OAuthService;
import dev.fitko.fitconnect.api.services.keys.KeyService;
import dev.fitko.fitconnect.api.services.submission.SubmissionService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.http.HttpEntity;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpMethod;
import org.springframework.http.MediaType;
import org.springframework.web.client.RestClientException;
import org.springframework.web.client.RestTemplate;

import java.nio.charset.StandardCharsets;
import java.text.ParseException;
import java.util.List;
import java.util.UUID;

public class PublicKeyService implements KeyService {

    private static final Logger LOGGER = LoggerFactory.getLogger(PublicKeyService.class);

    private static final ObjectMapper MAPPER = new ObjectMapper();

    private final ApplicationConfig config;
    private final ValidationService validationService;
    private final SubmissionService submissionService;
    private final RestTemplate restTemplate;
    private final OAuthService authService;

    public PublicKeyService(
            final ApplicationConfig config,
            final RestTemplate restTemplate,
            final OAuthService authService,
            final SubmissionService submissionService,
            final ValidationService validationService) {
        this.config = config;
        this.restTemplate = restTemplate;
        this.authService = authService;
        this.submissionService = submissionService;
        this.validationService = validationService;
    }

    @Override
    public RSAKey getPublicEncryptionKey(final UUID destinationId) {
        final Destination destination = submissionService.getDestination(destinationId);
        final String destinationUrl = config.getDestinationsKeyEndpoint();
        final ApiJwk publicKey = performRequest(destinationUrl, ApiJwk.class, getHeaders(), destinationId, destination.getEncryptionKid());
        final RSAKey rsaKey = toRSAKey(publicKey);
        validateEncryptionKey(rsaKey);
        return rsaKey;
    }

    @Override
    public RSAKey getPublicSignatureKey(final UUID destinationId, final String keyId) {
        final String destinationUrl = config.getDestinationsKeyEndpoint();
        final ApiJwk signatureKey = performRequest(destinationUrl, ApiJwk.class, getHeaders(), destinationId, keyId);
        final RSAKey rsaKey = toRSAKey(signatureKey);
        validateSignatureKey(rsaKey);
        return rsaKey;
    }

    @Override
    public RSAKey getSubmissionServicePublicKey(final String keyId) {
        final String submissionServiceUrl = config.getSubmissionServiceWellKnownKeysEndpoint();
        final ApiJwkSet wellKnownKeys = performRequest(submissionServiceUrl, ApiJwkSet.class, getHeaders());
        final RSAKey signatureKey = filterKeysById(keyId, wellKnownKeys.getKeys());
        validateSignatureKey(signatureKey);
        return signatureKey;
    }

    @Override
    public RSAKey getPortalPublicKey(final String keyId) {
        final String portalUrl = config.getSelfServicePortalWellKnownKeysEndpoint();
        final ApiJwkSet wellKnownKeys = performRequest(portalUrl, ApiJwkSet.class, getHeaders());
        final RSAKey signatureKey = filterKeysById(keyId, wellKnownKeys.getKeys());
        validateSignatureKey(signatureKey);
        return signatureKey;
    }
    @Override
    public RSAKey getWellKnownKeysForSubmissionUrl(final String url, final String keyId) {
        final var requestUrl = !url.endsWith("/") ? url + config.getWellKnownKeysPath() : url;
        final ApiJwkSet wellKnownKeys = performRequest(requestUrl, ApiJwkSet.class,  getHeadersWithoutAuth(), keyId);
        final RSAKey signatureKey = filterKeysById(keyId, wellKnownKeys.getKeys());
        validateSignatureKey(signatureKey);
        return signatureKey;
    }

    private RSAKey filterKeysById(final String keyId, final List<ApiJwk> keys) {
        return keys.stream()
                .filter(key -> key.getKid().equals(keyId))
                .map(this::toRSAKey)
                .findFirst()
                .orElse(null);
    }

    private void validateEncryptionKey(final RSAKey rsaKey) {
        final ValidationResult result = validationService.validateEncryptionPublicKey(rsaKey);
        validateResult(result, "Invalid public encryption key");
    }

    private void validateSignatureKey(final RSAKey rsaKey) {
        final ValidationResult result = validationService.validateSignaturePublicKey(rsaKey);
        validateResult(result, "Public signature key is not valid");
    }

    private void validateResult(final ValidationResult validationResult, final String message) {
        if (validationResult.hasError()) {
            if (config.getCurrentEnvironment().isAllowInsecurePublicKey()) {
                LOGGER.warn(message, validationResult.getError());
            } else {
                throw new InvalidKeyException(message, validationResult.getError());
            }
        }
    }

    private RSAKey toRSAKey(final ApiJwk jwk) {
        try {
            return RSAKey.parse(MAPPER.writeValueAsString(jwk));
        } catch (final JsonProcessingException | ParseException e) {
            throw new InvalidKeyException("Key could not be parsed", e);
        }
    }

    private <T> T performRequest(final String url, final Class<T> responseType, final HttpHeaders headers, final Object... params) {
        final HttpEntity<String> entity = new HttpEntity<>(headers);
        try {
            return restTemplate.exchange(url, HttpMethod.GET, entity, responseType, params).getBody();
        } catch (final RestClientException e) {
            throw new RestApiException("Request failed", e);
        }
    }

    private HttpHeaders getHeadersWithoutAuth() {
        final var headers = new HttpHeaders();
        headers.setContentType(MediaType.APPLICATION_JSON);
        headers.setAcceptCharset(List.of(StandardCharsets.UTF_8));
        return headers;
    }
    private HttpHeaders getHeaders() {
        final var headers = getHeadersWithoutAuth();
        headers.setBearerAuth(authService.getCurrentToken().getAccessToken());
        return headers;
    }
}
