package dev.fitko.fitconnect.core.schema;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import dev.fitko.fitconnect.api.config.SchemaConfig;
import dev.fitko.fitconnect.api.domain.schema.SchemaResources;
import dev.fitko.fitconnect.api.exceptions.InitializationException;
import dev.fitko.fitconnect.api.exceptions.SchemaNotFoundException;
import dev.fitko.fitconnect.api.services.schema.SchemaProvider;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URI;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

public class SchemaResourceProvider implements SchemaProvider {

    private static final Logger LOGGER = LoggerFactory.getLogger(SchemaResourceProvider.class);

    private static final Pattern ALLOWED_SCHEMA_PATTERN = Pattern.compile("1.\\d+.\\d+");

    private final Map<URI, String> setSchemas;

    private final Map<URI, String> metadataSchemas;

    private final Map<URI, String> destinationSchemas;

    private static final ObjectMapper MAPPER = new ObjectMapper();

    public SchemaResourceProvider(final SchemaResources schemaResources) {
        setSchemas = new HashMap<>();
        metadataSchemas = new HashMap<>();
        destinationSchemas = new HashMap<>();
        populateSetSchemas(schemaResources.getSetSchemaPaths());
        populateMetadataSchemas(schemaResources.getMetadataSchemaPaths());
        populateDestinationSchemas(schemaResources.getDestinationSchemaPaths());
    }

    private void populateMetadataSchemas(final List<String> metadataSchemaPaths) {
        LOGGER.info("Initializing metadata schemas");
        getResourceFiles(metadataSchemaPaths).forEach(this::addMetadataSchema);
    }

    private void populateSetSchemas(final List<String> setSchemaPaths) {
        LOGGER.info("Initializing set schemas");
        getResourceFiles(setSchemaPaths).forEach(this::addSetSchema);
    }

    private void populateDestinationSchemas(final List<String> destinationSchemaPaths) {
        LOGGER.info("Initializing destination schemas");
        getResourceFiles(destinationSchemaPaths).forEach(this::addDestinationSchema);
    }

    @Override
    public boolean isAllowedSetSchema(final URI schemaUri) {
        return schemaVersionMatchesPattern(schemaUri, ALLOWED_SCHEMA_PATTERN);
    }

    @Override
    public boolean isAllowedMetadataSchema(final URI schemaUri) {
        return metadataSchemas.containsKey(schemaUri);
    }

    @Override
    public String loadLatestSetSchema() throws SchemaNotFoundException {
        final URI latestSchema = SchemaConfig.SET_V_1_0_1.getSchemaUri();
        final String schema = setSchemas.get(latestSchema);
        if (schema == null) {
            throw new SchemaNotFoundException("SET schema " + latestSchema.toString() + " is not available.");
        }
        return schema;
    }

    @Override
    public String loadMetadataSchema(final URI schemaUri) throws SchemaNotFoundException {
        final String schema = metadataSchemas.get(schemaUri);
        if (schema == null) {
            throw new SchemaNotFoundException("Metadata schema " + schemaUri.toString() + " is not available.");
        }
        return schema;
    }

    @Override
    public String loadDestinationSchema(final URI schemaUri) throws SchemaNotFoundException {
        final String schema = destinationSchemas.get(schemaUri);
        if (schema == null) {
            throw new SchemaNotFoundException("Destination schema " + schemaUri.toString() + " is not available.");
        }
        return schema;
    }

    private void addSetSchema(final String schema) {
        setSchemas.put(readIdFromSchema(schema), schema);
    }

    private void addMetadataSchema(final String schema) {
        metadataSchemas.put(readIdFromSchema(schema), schema);
    }

    private void addDestinationSchema(final String schema) {
        destinationSchemas.put(readIdFromSchema(schema), schema);
    }

    private URI readIdFromSchema(final String schema) {
        try {
            return URI.create(MAPPER.readTree(schema).get("$id").asText());
        } catch (final JsonProcessingException e) {
            throw new InitializationException("Parsing schema failed", e);
        }
    }

    private List<String> getResourceFiles(final List<String> schemaFiles) {
        return schemaFiles.stream().map(SchemaResourceProvider::readResourceToString).collect(Collectors.toList());
    }

    private static String readResourceToString(final String schemaFile) {
        try (final InputStream is = SchemaResourceProvider.class.getResourceAsStream(schemaFile);
             final BufferedReader reader = new BufferedReader(new InputStreamReader(is))) {
            return reader.lines().collect(Collectors.joining(System.lineSeparator()));
        } catch (final Exception e) {
            throw new InitializationException("Could not read schema file " + schemaFile, e);
        }
    }

    private boolean schemaVersionMatchesPattern(final URI schemaUri, final Pattern pattern) {
        final String schemaVersion = schemaUri.getPath().split("/")[3];
        return pattern.matcher(schemaVersion).matches();
    }
}
