package dev.fitko.fitconnect.core.util;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jwt.JWTClaimsSet;
import com.nimbusds.jwt.SignedJWT;
import dev.fitko.fitconnect.api.domain.model.event.*;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.event.problems.Problem;
import dev.fitko.fitconnect.api.exceptions.EventCreationException;
import dev.fitko.fitconnect.api.exceptions.EventLogException;

import java.text.ParseException;
import java.util.*;
import java.util.function.Predicate;
import java.util.stream.Collectors;

import static com.fasterxml.jackson.databind.DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES;
import static dev.fitko.fitconnect.api.domain.model.event.EventIssuer.*;

public final class EventLogUtil {

    private static final ObjectMapper MAPPER = new ObjectMapper().configure(FAIL_ON_UNKNOWN_PROPERTIES, false);

    private EventLogUtil() {
    }

    public static List<EventLogEntry> mapEventLogToEntries(final EventLog eventLog) {
        return eventLog.getEventLogs()
                .stream()
                .map(EventLogUtil::eventToEntry)
                .collect(Collectors.toList());
    }

    public static EventIssuer resolveIssuerType(final String issuer) {
        if (issuer == null) {
            return UNDEFINED;
        }
        return issuer.startsWith("http") ? SUBMISSION_SERVICE : DESTINATION;
    }

    public static Event getEventFromClaims(final JWTClaimsSet claims) {
        final Map<String, Object> events = (Map) claims.getClaim(EventClaimFields.CLAIM_EVENTS);
        final Map.Entry<String, Map> eventEntry = (Map.Entry) events.entrySet().iterator().next();
        return Event.fromSchemaUri(eventEntry.getKey());
    }

    public static Event getEventFromJWT(final SignedJWT event) {
        try {
            final JWTClaimsSet claims = event.getJWTClaimsSet();
            final Map<String, Object> events = (Map) claims.getClaim(EventClaimFields.CLAIM_EVENTS);
            final Map.Entry<String, Map> eventEntry = (Map.Entry) events.entrySet().iterator().next();
            return Event.fromSchemaUri(eventEntry.getKey());
        } catch (final ParseException e) {
            throw new EventLogException(e.getMessage(), e);
        }
    }

    public static UUID getDestinationId(final String issuer) {
        final EventIssuer issuerType = resolveIssuerType(issuer);
        if (issuerType.equals(SUBMISSION_SERVICE) || issuerType.equals(UNDEFINED)) {
            return null;
        }
        try {
            return UUID.fromString(issuer);
        } catch (final IllegalArgumentException ex) {
            throw new EventLogException("Destination id '" + issuer + "' from issuer is no valid uuid");
        }
    }

    private static EventLogEntry eventToEntry(final String setEvent) {
        return eventToEntry(parseEvent(setEvent));
    }

    public static EventLogEntry eventToEntry(final SignedJWT signedJWT) {
        final JWTClaimsSet claimsSet = getClaimsSet(signedJWT);
        final Map<String, Object> events = (Map) claimsSet.getClaim(EventClaimFields.CLAIM_EVENTS);
        final Map.Entry<String, Map> eventEntry = (Map.Entry) events.entrySet().iterator().next();
        final String eventName = eventEntry.getKey();
        final Map<String, Object> eventPayload = eventEntry.getValue();
        final List<Problem> problems = new ArrayList<>();
        if (eventPayload.containsKey("problems")) {
            problems.addAll((List<Problem>) eventPayload.get("problems"));
        }
        return EventLogEntry.builder()
                .event(Event.fromSchemaUri(eventName))
                .issuer(claimsSet.getIssuer())
                .issueTime(claimsSet.getIssueTime())
                .eventId(getEventId(claimsSet))
                .caseId(getTransactionClaim(claimsSet))
                .submissionId(getSubmissionId(claimsSet))
                .problems(problems)
                .build();
    }

    public static List<SignedJWT> getJWTSFromEvents(final List<String> events) {
        return events.stream()
                .map(EventLogUtil::getJWTFromEvent)
                .collect(Collectors.toList());
    }

    private static SignedJWT parseEvent(final String event) {
        try {
            return SignedJWT.parse(event);
        } catch (final ParseException e) {
            throw new EventLogException(e.getMessage(), e);
        }
    }

    private static UUID getSubmissionId(final JWTClaimsSet claimsSet) {
        return TypeAndUUID.fromString(claimsSet.getSubject()).getUuid();
    }

    private static UUID getEventId(final JWTClaimsSet claimsSet) {
        return UUID.fromString(claimsSet.getJWTID());
    }

    private static UUID getTransactionClaim(final JWTClaimsSet claimsSet) {
        try {
            return TypeAndUUID.fromString(claimsSet.getStringClaim(EventClaimFields.CLAIM_TXN)).getUuid();
        } catch (final ParseException e) {
            throw new EventLogException(e.getMessage(), e);
        }
    }

    private static JWTClaimsSet getClaimsSet(final SignedJWT signedJWT) {
        try {
            return signedJWT.getJWTClaimsSet();
        } catch (final ParseException e) {
            throw new EventCreationException(e.getMessage(), e);
        }
    }

    private static SignedJWT getJWTFromEvent(final String event) {
        try {
            return SignedJWT.parse(event);
        } catch (final ParseException e) {
            throw new EventCreationException(e.getMessage(), e);
        }
    }

    public static AuthenticationTags getAuthenticationTags(final JWTClaimsSet claims) {
        try {
            final Map<String, Object> eventsClaim = (Map) claims.getClaim(EventClaimFields.CLAIM_EVENTS);
            final Map.Entry<String, Map> eventEntry = (Map.Entry) eventsClaim.entrySet().iterator().next();
            final Map<String, Object> events = eventEntry.getValue();
            final String authTags = MAPPER.writeValueAsString(events.get("authenticationTags"));
            return MAPPER.readValue(authTags, AuthenticationTags.class);
        } catch (final JsonProcessingException e) {
            throw new EventCreationException(e.getMessage(), e);
        }
    }

    public static Predicate<SignedJWT> eventEqualsSubmissionId(final UUID submissionId) {
        return jwt -> {
            try {
                final String subject = jwt.getJWTClaimsSet().getSubject();
                return TypeAndUUID.fromString(subject).getUuid().equals(submissionId);
            } catch (final ParseException e) {
                throw new EventCreationException(e.getMessage(), e);
            }
        };
    }
}
