package dev.fitko.fitconnect.core.crypto;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.jwk.KeyUse;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jose.jwk.gen.RSAKeyGenerator;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import org.junit.jupiter.api.Test;

import java.nio.charset.StandardCharsets;
import java.util.Arrays;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.*;

public class JWECryptoServiceTest {

    final CryptoService underTest = new JWECryptoService(new HashService());

    @Test
    void encryptString() throws JOSEException {
        // Given
        final RSAKey key = generateRsaKey(4096);
        final RSAKey publicKey = key.toPublicJWK();

        // When
        final String encryptedData = underTest.encryptString(publicKey, "some foo");

        // Then
        assertNotNull(encryptedData);
        assertNotEquals("some foo", encryptedData);
    }

    @Test
    void decryptString() throws JOSEException {
        // Given
        final RSAKey privateKey = generateRsaKey(4096);
        final RSAKey publicKey = privateKey.toPublicJWK();
        final String dataFromSender = underTest.encryptString(publicKey, "some foo");

        // When
        final String decrypted = underTest.decryptString(privateKey, dataFromSender);

        // Then
        assertNotNull(decrypted);
        assertEquals("some foo", decrypted);
    }

    @Test
    void encryptBytes() throws JOSEException {
        // Given
        final RSAKey key = generateRsaKey(4096);
        final RSAKey publicKey = key.toPublicJWK();

        // When
        final String encryptedData = underTest.encryptBytes(publicKey, "some foo".getBytes(StandardCharsets.UTF_8));

        // Then
        assertNotNull(encryptedData);
        assertNotEquals("some foo", encryptedData);
    }

    @Test
    void encryptObject() throws JOSEException, JsonProcessingException {
        // Given
        final RSAKey key = generateRsaKey(4096);

        final Metadata metadata = new Metadata();

        // When
        final String encryptedData = underTest.encryptObject(key, metadata);

        // Then
        assertNotNull(encryptedData);
        assertThat(new ObjectMapper().readValue(underTest.decryptString(key, encryptedData), Metadata.class), is(metadata));
    }


    @Test
    void decryptBytes() throws JOSEException {
        // Given
        final RSAKey privateKey = generateRsaKey(4096);
        final RSAKey publicKey = privateKey.toPublicJWK();
        final String dataFromSender = underTest.encryptString(publicKey, "some foo");

        // When
        final byte[] decrypted = underTest.decryptBytes(privateKey, dataFromSender);

        // Then
        assertNotNull(decrypted);
        assertArrayEquals(decrypted, "some foo".getBytes(StandardCharsets.UTF_8));
    }


    @Test
    void hashBytes() {
        final var expectedHash = "6cb40f8c9a1f4411c6b83fefdd950009ec2906e3747e9146e7704234cdb6a7b11eefb9e6221fdad3cda7c5ece6a62ef535ca2ac84ba61226c9835062cf84db56";
        final byte[] contentToHash = "content to hash".getBytes(StandardCharsets.UTF_8);
        final String sha512Hash = underTest.hashBytes(contentToHash);

        assertEquals(expectedHash, sha512Hash);
        assertEquals(underTest.hashBytes(contentToHash), underTest.hashBytes(contentToHash));
    }

    public static RSAKey generateRsaKey(final int size) throws JOSEException {
        return new RSAKeyGenerator(size)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate();
    }
}