package dev.fitko.fitconnect.core.events;

import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.JWSAlgorithm;
import com.nimbusds.jose.JWSVerifier;
import com.nimbusds.jose.crypto.RSASSAVerifier;
import com.nimbusds.jose.jwk.JWK;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jwt.SignedJWT;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.config.Environment;
import dev.fitko.fitconnect.api.config.EnvironmentName;
import dev.fitko.fitconnect.api.config.SchemaConfig;
import dev.fitko.fitconnect.api.domain.model.event.Event;
import dev.fitko.fitconnect.api.domain.model.event.EventPayload;
import dev.fitko.fitconnect.api.domain.model.event.problems.submission.AttachmentsMismatch;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.schema.SchemaResources;
import dev.fitko.fitconnect.api.exceptions.EventCreationException;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.api.services.events.SecurityEventService;
import dev.fitko.fitconnect.api.services.schema.SchemaProvider;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import dev.fitko.fitconnect.core.crypto.HashService;
import dev.fitko.fitconnect.core.crypto.JWECryptoService;
import dev.fitko.fitconnect.core.schema.SchemaResourceProvider;
import dev.fitko.fitconnect.core.util.FileUtil;
import dev.fitko.fitconnect.core.validation.DefaultValidationService;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.text.ParseException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.*;

class SecurityEventTokenServiceTest {

    private RSAKey signingKey;
    private RSAKey encryptionKey;
    private CryptoService cryptoService;
    private ValidationService validationService;

    private SecurityEventService underTest;

    @BeforeEach
    void setUp() throws IOException, ParseException {
        final ApplicationConfig config = getApplicationConfig();
        cryptoService = new JWECryptoService(new HashService());
        signingKey = JWK.parse(getResourceAsString("private_test_signing_key.json")).toRSAKey();
        encryptionKey = JWK.parse(getResourceAsString("public_encryption_test_key.json")).toRSAKey();

        final List<String> setSchemas = SchemaConfig.getSetSchemaFilePaths("/set-schema");
        final List<String> metadataSchemas = SchemaConfig.getMetadataSchemaFileNames("/metadata-schema");
        final List<String> destinationSchemas = SchemaConfig.getDestinationSchemaPaths("/destination-schema");
        final SchemaResources schemaResources = new SchemaResources(setSchemas, metadataSchemas, destinationSchemas);
        final SchemaProvider schemaProvider = new SchemaResourceProvider(schemaResources);
        this.validationService = new DefaultValidationService(config, new HashService(), schemaProvider,
                FileUtil.loadContentOfFilesInDirectory("trusted-test-root-certificates"));
        this.underTest = new SecurityEventTokenService(config, this.validationService, this.signingKey);
    }

    @Test
    void testAcceptSubmissionEventWithoutProblems() throws JOSEException {

        // Given
        final var encryptedData = cryptoService.encryptString(encryptionKey, "test data");
        final var encryptedMetadata = cryptoService.encryptString(encryptionKey, "test metadata");

        final var encryptedAttachments = new HashMap<UUID, String>();
        encryptedAttachments.put(UUID.randomUUID(), cryptoService.encryptString(encryptionKey, "test attachment 1"));
        encryptedAttachments.put(UUID.randomUUID(), cryptoService.encryptString(encryptionKey, "test attachment 2"));

        final var submission = new Submission();
        submission.setSubmissionId(UUID.randomUUID());
        submission.setDestinationId(UUID.randomUUID());
        submission.setCaseId(UUID.randomUUID());
        submission.setEncryptedData(encryptedData);
        submission.setEncryptedMetadata(encryptedMetadata);

        final EventPayload eventPayload = new EventPayload(submission, encryptedAttachments);

        // When
        final SignedJWT signedJWT = underTest.createAcceptSubmissionEvent(eventPayload);

        // Then
        final Map<String, Object> payload = signedJWT.getPayload().toJSONObject();
        final var eventsClaim = (Map) payload.get("events");
        final var eventObject = (Map) eventsClaim.get(Event.ACCEPT.getSchemaUri());
        final JWSVerifier jwsVerifier = new RSASSAVerifier(signingKey);

        assertEquals(JWSAlgorithm.PS512, signingKey.getAlgorithm());
        assertTrue(signedJWT.verify(jwsVerifier));
        assertThat(payload.get("sub"), is("submission:" + eventPayload.getSubmissionId()));
        assertThat(payload.get("txn"), is("case:" + eventPayload.getCaseId()));
        assertThat(payload.get("iss"), is(eventPayload.getDestinationId().toString()));
        assertThat(eventObject.size(), is(1));
    }

    @Test
    void testAcceptSubmissionEventWithProblem() throws JOSEException {

        // Given
        final var encryptedData = cryptoService.encryptString(encryptionKey, "test data");
        final var encryptedMetadata = cryptoService.encryptString(encryptionKey, "test metadata");

        final var encryptedAttachments = new HashMap<UUID, String>();
        encryptedAttachments.put(UUID.randomUUID(), cryptoService.encryptString(encryptionKey, "test attachment"));

        final var submission = new Submission();
        submission.setSubmissionId(UUID.randomUUID());
        submission.setDestinationId(UUID.randomUUID());
        submission.setCaseId(UUID.randomUUID());
        submission.setEncryptedData(encryptedData);
        submission.setEncryptedMetadata(encryptedMetadata);

        final var problem = new AttachmentsMismatch();

        final EventPayload eventPayload = new EventPayload(submission, encryptedAttachments, List.of(problem));

        // When
        final SignedJWT signedJWT = underTest.createAcceptSubmissionEvent(eventPayload);

        // Then
        final Map<String, Object> payload = signedJWT.getPayload().toJSONObject();
        final var eventsClaim = (Map) payload.get("events");
        final var eventObject = (Map) eventsClaim.get(Event.ACCEPT.getSchemaUri());
        final JWSVerifier jwsVerifier = new RSASSAVerifier(signingKey);

        assertEquals(JWSAlgorithm.PS512, signingKey.getAlgorithm());
        assertTrue(signedJWT.verify(jwsVerifier));
        assertThat(payload.get("sub"), is("submission:" + eventPayload.getSubmissionId()));
        assertThat(payload.get("txn"), is("case:" + eventPayload.getCaseId()));
        assertThat(payload.get("iss"), is(eventPayload.getDestinationId().toString()));
        assertThat(eventObject.size(), is(2));
        assertThat(((List) eventObject.get("problems")).get(0), is(problem));
    }

    @Test
    void testRejectSubmissionEvent() throws JOSEException {

        // Given
        final var problem = new AttachmentsMismatch();

        final var submission = new Submission();
        submission.setSubmissionId(UUID.randomUUID());
        submission.setDestinationId(UUID.randomUUID());
        submission.setCaseId(UUID.randomUUID());

        final EventPayload eventPayload = new EventPayload(submission, List.of(problem));

        // When
        final SignedJWT signedJWT = underTest.createRejectSubmissionEvent(eventPayload);

        // Then
        final Map<String, Object> payload = signedJWT.getPayload().toJSONObject();
        final var eventsClaim = (Map) payload.get("events");
        final var eventObject = (Map) eventsClaim.get(Event.REJECT.getSchemaUri());
        final JWSVerifier jwsVerifier = new RSASSAVerifier(signingKey);

        assertEquals(JWSAlgorithm.PS512, signingKey.getAlgorithm());
        assertTrue(signedJWT.verify(jwsVerifier));
        assertThat(payload.get("sub"), is("submission:" + eventPayload.getSubmissionId()));
        assertThat(payload.get("txn"), is("case:" + eventPayload.getCaseId()));
        assertThat(payload.get("iss"), is(eventPayload.getDestinationId().toString()));
        assertThat(eventObject.size(), is(1));
        assertThat(((List) eventObject.get("problems")).get(0), is(problem));
    }

    @Test
    void testAcceptEventWithMissingMetadataAuthTag() {

        // Given
        final var submission = new Submission();
        submission.setSubmissionId(UUID.randomUUID());
        submission.setDestinationId(UUID.randomUUID());
        submission.setCaseId(UUID.randomUUID());
        submission.setEncryptedData(cryptoService.encryptString(encryptionKey, "test data"));

        final EventPayload eventPayload = new EventPayload(submission, Map.of());

        // When
        final EventCreationException exception = assertThrows(EventCreationException.class, () -> underTest.createAcceptSubmissionEvent(eventPayload));

        // Then
        assertThat(exception.getCause().getMessage(), containsString("$.events.https://schema.fitko.de/fit-connect/events/accept-submission.authenticationTags.metadata: null found"));
    }

    @Test
    void testAcceptEventWithMissingDataAuthTag() {

        // Given
        final var submission = new Submission();
        submission.setSubmissionId(UUID.randomUUID());
        submission.setDestinationId(UUID.randomUUID());
        submission.setCaseId(UUID.randomUUID());
        submission.setEncryptedMetadata(cryptoService.encryptString(encryptionKey, "test metadata"));

        final EventPayload eventPayload = new EventPayload(submission, Map.of());

        // When
        final EventCreationException exception = assertThrows(EventCreationException.class, () -> underTest.createAcceptSubmissionEvent(eventPayload));

        // Then
        assertThat(exception.getCause().getMessage(), containsString("$.events.https://schema.fitko.de/fit-connect/events/accept-submission.authenticationTags.data: null found"));
    }

    @Test
    void testAcceptEventWithMissingAuthData() {

        // Given
        final var submission = new Submission();
        submission.setSubmissionId(UUID.randomUUID());
        submission.setDestinationId(UUID.randomUUID());
        submission.setCaseId(UUID.randomUUID());

        final EventPayload eventPayload = new EventPayload(submission);

        // When
        final EventCreationException exception = assertThrows(EventCreationException.class, () -> underTest.createAcceptSubmissionEvent(eventPayload));

        // Then
        assertThat(exception.getCause().getMessage(), containsString("$.events.https://schema.fitko.de/fit-connect/events/accept-submission.authenticationTags.metadata: null found"));
        assertThat(exception.getCause().getMessage(), containsString("$.events.https://schema.fitko.de/fit-connect/events/accept-submission.authenticationTags.data: null found"));
    }

    @Test
    void testRejectEventWithMissingProblems() {

        // Given
        final var submission = new Submission();
        submission.setSubmissionId(UUID.randomUUID());
        submission.setDestinationId(UUID.randomUUID());
        submission.setCaseId(UUID.randomUUID());

        final EventPayload eventPayload = new EventPayload(submission);

        // When
        final EventCreationException exception = assertThrows(EventCreationException.class, () -> underTest.createRejectSubmissionEvent(eventPayload));

        // Then
        assertThat(exception.getCause().getMessage(), containsString("$.events.https://schema.fitko.de/fit-connect/events/reject-submission.problems: is missing but it is required"));
    }

    private static ApplicationConfig getApplicationConfig() {
        final var envName = new EnvironmentName("testing");
        final var testing = new Environment();
        testing.setAllowInsecurePublicKey(true);
        final var config = new ApplicationConfig();
        config.setEnvironments(Map.of(envName, testing));
        config.setActiveEnvironment(envName);
        return config;
    }

    private String getResourceAsString(final String name) throws IOException {
        final ClassLoader classLoader = getClass().getClassLoader();
        return Files.readString(new File(classLoader.getResource(name).getFile()).toPath());
    }
}