package dev.fitko.fitconnect.core.http;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import dev.fitko.fitconnect.api.config.BuildInfo;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.SignatureType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.core.RestEndpointBase;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.springframework.http.MediaType;
import org.springframework.http.converter.HttpMessageConverter;
import org.springframework.http.converter.json.MappingJackson2HttpMessageConverter;
import org.springframework.web.client.RestTemplate;

import java.net.URI;
import java.util.Collections;
import java.util.List;
import java.util.Map;

import static com.github.tomakehurst.wiremock.client.WireMock.get;
import static com.github.tomakehurst.wiremock.client.WireMock.ok;
import static com.github.tomakehurst.wiremock.client.WireMock.urlEqualTo;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.greaterThan;
import static org.hamcrest.Matchers.instanceOf;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertNotNull;

public class RestTemplateTest extends RestEndpointBase {

    private RestTemplate underTest;

    @BeforeEach
    void setup() {
        underTest = new RestService(null, 0, new BuildInfo()).getRestTemplate();
    }

    @Test
    void testRestCallForUnknownProperties() throws JsonProcessingException {

        // Given
        final var fakeBaseUrl = "http://localhost:" + wireMockServer.port();

        final Map<String, Map<String, Object>> metadataWithUnknownProperty = getMetadataWithUnknownSchemaProperty();

        wireMockServer.stubFor(
                get(urlEqualTo("/test"))
                        .willReturn(ok()
                                .withBody(new ObjectMapper().writeValueAsString(metadataWithUnknownProperty))
                                .withHeader("Content-Type", MediaType.APPLICATION_JSON_VALUE)
                                .withStatus(200)));
        
        // When
        final Metadata response = underTest.getForObject(URI.create(fakeBaseUrl + "/test"), Metadata.class);

        // Then
        assertNotNull(response);
    }

    @Test
    void testMessageConversionForUnknownProperties() throws JsonProcessingException {

        // Given
        final List<HttpMessageConverter<?>> messageConverters = underTest.getMessageConverters();
        assertThat(messageConverters.size(), is(greaterThan(0)));

        // Get last set converter that's expected to be the jackson object mapper
        final HttpMessageConverter<?> httpMessageConverter = messageConverters.get(messageConverters.size()-1);
        assertThat(httpMessageConverter, instanceOf(MappingJackson2HttpMessageConverter.class));

        final MappingJackson2HttpMessageConverter converter = (MappingJackson2HttpMessageConverter) httpMessageConverter;

        final ObjectMapper mapper = converter.getObjectMapper();

        final Map<String, Map<String, Object>> metadataWithUnknownProperty = getMetadataWithUnknownSchemaProperty();

        // Map object with an unknown property
        final Metadata metadata = mapper.readValue(mapper.writeValueAsString(metadataWithUnknownProperty), Metadata.class);

        assertNotNull(metadata);
    }

    private static Map<String, Map<String, Object>> getMetadataWithUnknownSchemaProperty() {
        return Map.of(
                "$schemaFoo", Collections.emptyMap(),
                "contentStructure", Map.of(
                        "data", Map.of(
                                "hash", Map.of(
                                        "type", SignatureType.SHA_512,
                                        "content", "bla"
                                ),
                                "submissionSchema", Map.of(
                                        "schemaUri", URI.create("https://dummy.schema.url"),
                                        "mimeType", MimeType.APPLICATION_JSON
                                )
                        )
                )
        );
    }
}
