package dev.fitko.fitconnect.core.routing;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.JWSAlgorithm;
import com.nimbusds.jose.JWSHeader;
import com.nimbusds.jose.Payload;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jose.util.Base64URL;
import com.nimbusds.jwt.SignedJWT;
import dev.fitko.fitconnect.api.domain.model.route.Route;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.ValidationException;
import dev.fitko.fitconnect.api.services.keys.KeyService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import dev.fitko.fitconnect.core.SubmissionSenderTest;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.text.ParseException;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.anyMap;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

class RouteVerifierTest {

    private RouteVerifier underTest;
    private KeyService keyServiceMock;
    private ValidationService validatorMock;
    private RSAKey portalSignatureKey;
    private RSAKey submissionServiceSignatureKey;

    @BeforeEach
    void setup() throws IOException, ParseException {
        keyServiceMock = mock(KeyService.class);
        validatorMock = mock(ValidationService.class);
        underTest = new RouteVerifier(keyServiceMock, validatorMock);
        portalSignatureKey = RSAKey.parse(getResourceAsString("/portal_well_known_signature_key.json"));
        submissionServiceSignatureKey = RSAKey.parse(getResourceAsString("/submission_service_well_known_signature_key.json"));
    }

    @Test
    void validateRouteDestinationTest() throws IOException {

        // Given
        final Route route = new ObjectMapper().readValue(getResourceAsString("/sample_route.json"), Route.class);

        when(keyServiceMock.getPortalPublicKey(anyString())).thenReturn(portalSignatureKey);
        when(keyServiceMock.getWellKnownKeysForSubmissionUrl(anyString(), anyString())).thenReturn(submissionServiceSignatureKey);
        when(validatorMock.validateDestinationSchema(anyMap())).thenReturn(ValidationResult.ok());


        // When
        final ValidationResult validationResult = underTest.validateRouteDestinations(List.of(route), "urn:de:fim:leika:leistung:99123456760610", "064350014014");

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    void hostUrlDoesNotMatchTest() throws IOException {

        // Given
        final Route route = new ObjectMapper().readValue(getResourceAsString("/sample_route.json"), Route.class);
        route.getDestinationParameters().setSubmissionUrl("https://dev.null.net");

        when(keyServiceMock.getPortalPublicKey(anyString())).thenReturn(portalSignatureKey);
        when(keyServiceMock.getWellKnownKeysForSubmissionUrl(anyString(), anyString())).thenReturn(submissionServiceSignatureKey);
        when(validatorMock.validateDestinationSchema(anyMap())).thenReturn(ValidationResult.ok());

        // When
        final ValidationResult validationResult = underTest.validateRouteDestinations(List.of(route), "urn:de:fim:leika:leistung:99123456760610", "064350014014");

        // Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("Submission host does not match destinationParameters submission url submission-api-testing.fit-connect.fitko.dev"));
    }

    @Test
    void requiredAlgorithmDoesNotMatchTest() throws IOException, ParseException {

        // Given
        final Route route = new ObjectMapper().readValue(getResourceAsString("/sample_route.json"), Route.class);
        final SignedJWT signature = SignedJWT.parse(route.getDestinationSignature());

        final Base64URL headerWithWrongAlgorithm = new JWSHeader(JWSAlgorithm.ES512).toBase64URL();
        final SignedJWT signatureWithWrongAlg = new SignedJWT(headerWithWrongAlgorithm, signature.getParsedParts()[1], signature.getParsedParts()[2]);
        route.setDestinationSignature(signatureWithWrongAlg.serialize());

        when(keyServiceMock.getPortalPublicKey(anyString())).thenReturn(portalSignatureKey);
        when(keyServiceMock.getWellKnownKeysForSubmissionUrl(anyString(), anyString())).thenReturn(submissionServiceSignatureKey);

        // When
        final ValidationResult validationResult = underTest.validateRouteDestinations(List.of(route), "urn:de:fim:leika:leistung:99123456760610", "064350014014");

        // Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("Algorithm in signature header is not PS512"));
    }

    @Test
    void requiredClaimsNotPresentTest() throws IOException, ParseException {

        // Given
        final Route route = new ObjectMapper().readValue(getResourceAsString("/sample_route.json"), Route.class);
        final SignedJWT signature = SignedJWT.parse(route.getDestinationSignature());

        final Payload payloadWithMissingClaims = new Payload(Map.of("submissionHost", "submission-api-testing.fit-connect.fitko.dev"));
        final SignedJWT signatureWithMissingClaims = new SignedJWT(signature.getHeader().toBase64URL(), payloadWithMissingClaims.toBase64URL(), signature.getParsedParts()[2]);

        route.setDestinationSignature(signatureWithMissingClaims.serialize());

        when(keyServiceMock.getPortalPublicKey(anyString())).thenReturn(portalSignatureKey);
        when(keyServiceMock.getWellKnownKeysForSubmissionUrl(anyString(), anyString())).thenReturn(submissionServiceSignatureKey);
        when(validatorMock.validateDestinationSchema(anyMap())).thenReturn(ValidationResult.error(new ValidationException("Payload missing required claims")));


        // When
        final ValidationResult validationResult = underTest.validateRouteDestinations(List.of(route), "urn:de:fim:leika:leistung:99123456760610", "064350014014");

        // Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("Payload missing required claims"));
    }

    @Test
    void requestedRegionNotSupportedByDestinationTest() throws IOException, ParseException {

        // Given
        final Route route = new ObjectMapper().readValue(getResourceAsString("/sample_route.json"), Route.class);
        final SignedJWT signature = SignedJWT.parse(route.getDestinationSignature());

        final List<Map<String, List<String>>> serviceClaim = List.of(
                Map.of(
                        "gebietIDs", List.of("1234567"),
                        "leistungIDs", List.of("urn:de:fim:leika:leistung:99123456760610")
                )
        );

        final Map<String, Object> claims = Map.of(
                "submissionHost", "submission-api-testing.fit-connect.fitko.dev",
                "iss", "fit-connect",
                "iat", new Date().getTime(),
                "jti", UUID.randomUUID().toString(),
                "destinationId", UUID.randomUUID().toString(),
                "services", serviceClaim);

        final Payload payloadWithServicesNotMatching = new Payload(claims);
        final SignedJWT signatureWithServices = new SignedJWT(signature.getHeader().toBase64URL(), payloadWithServicesNotMatching.toBase64URL(), signature.getParsedParts()[2]);

        route.setDestinationSignature(signatureWithServices.serialize());

        when(keyServiceMock.getPortalPublicKey(anyString())).thenReturn(portalSignatureKey);
        when(keyServiceMock.getWellKnownKeysForSubmissionUrl(anyString(), anyString())).thenReturn(submissionServiceSignatureKey);
        when(validatorMock.validateDestinationSchema(anyMap())).thenReturn(ValidationResult.ok());


        // When
        final ValidationResult validationResult = underTest.validateRouteDestinations(List.of(route), "urn:de:fim:leika:leistung:99123456760610", "064350014014");

        // Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("Requested region '064350014014' is not supported by the destinations services"));
    }

    @Test
    void requestedServiceIdentifierNotSupportedByDestinationTest() throws IOException, ParseException {

        // Given
        final Route route = new ObjectMapper().readValue(getResourceAsString("/sample_route.json"), Route.class);
        final SignedJWT signature = SignedJWT.parse(route.getDestinationSignature());

        final List<Map<String, List<String>>> serviceClaim = List.of(
                Map.of(
                        "gebietIDs", List.of("064350014014"),
                        "leistungIDs", List.of("urn:de:fim:leika:leistung:1111111111111")
                )
        );

        final Map<String, Object> claims = Map.of(
                "submissionHost", "submission-api-testing.fit-connect.fitko.dev",
                "iss", "fit-connect",
                "iat", new Date().getTime(),
                "jti", UUID.randomUUID().toString(),
                "destinationId", UUID.randomUUID().toString(),
                "services", serviceClaim);

        final Payload payloadWithServicesNotMatching = new Payload(claims);
        final SignedJWT signatureWithServices = new SignedJWT(signature.getHeader().toBase64URL(), payloadWithServicesNotMatching.toBase64URL(), signature.getParsedParts()[2]);

        route.setDestinationSignature(signatureWithServices.serialize());

        when(keyServiceMock.getPortalPublicKey(anyString())).thenReturn(portalSignatureKey);
        when(keyServiceMock.getWellKnownKeysForSubmissionUrl(anyString(), anyString())).thenReturn(submissionServiceSignatureKey);
        when(validatorMock.validateDestinationSchema(anyMap())).thenReturn(ValidationResult.ok());

        // When
        final ValidationResult validationResult = underTest.validateRouteDestinations(List.of(route), "urn:de:fim:leika:leistung:99123456760610", "064350014014");

        // Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("Requested service identifier 'urn:de:fim:leika:leistung:99123456760610' is not supported by the destinations services"));
    }

    @Test
    void signatureVerificationFailedTest() throws IOException {

        // Given
        final Route route = new ObjectMapper().readValue(getResourceAsString("/sample_route.json"), Route.class);
        route.getDestinationParameters().setEncryptionKid("Wrong_Key_Id_That_Fails_The_Public_Key_Retrieval");

        when(keyServiceMock.getPortalPublicKey(anyString())).thenReturn(portalSignatureKey);
        when(keyServiceMock.getWellKnownKeysForSubmissionUrl(anyString(), anyString())).thenReturn(submissionServiceSignatureKey);
        when(validatorMock.validateDestinationSchema(anyMap())).thenReturn(ValidationResult.ok());


        // When
        final ValidationResult validationResult = underTest.validateRouteDestinations(List.of(route), "urn:de:fim:leika:leistung:99123456760610", "064350014014");

        // Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("Invalid destination parameter signature for route d40e7b13-da98-4b09-9e16-bbd61ca81510"));
    }

    @Test
    void combiningDetachedSignatureWithPayloadFailedTest() throws IOException {

        // Given
        final Route route = new ObjectMapper().readValue(getResourceAsString("/sample_route.json"), Route.class);

        // replace detached signature payload part with wrong data  (..)
       route.setDestinationParametersSignature(route.getDestinationParametersSignature().replace("..", ""));

        when(keyServiceMock.getPortalPublicKey(anyString())).thenReturn(portalSignatureKey);
        when(keyServiceMock.getWellKnownKeysForSubmissionUrl(anyString(), anyString())).thenReturn(submissionServiceSignatureKey);
        when(validatorMock.validateDestinationSchema(anyMap())).thenReturn(ValidationResult.ok());


        // When
        final ValidationResult validationResult = underTest.validateRouteDestinations(List.of(route), "urn:de:fim:leika:leistung:99123456760610", "064350014014");

        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("Invalid serialized unsecured/JWS/JWE object: Missing part delimiters"));
    }

    private String getResourceAsString(final String filename) throws IOException {
        return new String(SubmissionSenderTest.class.getResourceAsStream(filename).readAllBytes());
    }
}