package dev.fitko.fitconnect.core.schema;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import dev.fitko.fitconnect.api.config.SchemaConfig;
import dev.fitko.fitconnect.api.domain.schema.SchemaResources;
import dev.fitko.fitconnect.api.exceptions.SchemaNotFoundException;
import dev.fitko.fitconnect.api.services.schema.SchemaProvider;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.net.URI;
import java.util.List;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.equalTo;
import static org.junit.jupiter.api.Assertions.*;

class SchemaResourceProviderTest {

    private SchemaProvider underTest;

    @BeforeEach
    void setup() {
        final List<String> setSchemas = SchemaConfig.getSetSchemaFilePaths("/set-schema");
        final List<String> metadataSchemas = SchemaConfig.getMetadataSchemaFileNames("/metadata-schema");
        final List<String> destinationSchemas = SchemaConfig.getDestinationSchemaPaths("/destination-schema");
        final SchemaResources schemaResources = new SchemaResources(setSchemas, metadataSchemas, destinationSchemas);
        underTest = new SchemaResourceProvider(schemaResources);
    }

    @Test
    void isAllowedSetSchema() {
        assertTrue(underTest.isAllowedSetSchema(SchemaConfig.SET_V_1_0_1.getSchemaUri()));
        assertTrue(underTest.isAllowedSetSchema(SchemaConfig.SET_V_1_0_0.getSchemaUri()));
        assertTrue(underTest.isAllowedSetSchema(URI.create("https://schema.fitko.de/fit-connect/set-payload/1.2.3/set-payload.schema.json")));
    }

    @Test
    void isNoAllowedSetSchema() {
        assertFalse(underTest.isAllowedSetSchema(URI.create("https://schema.fitko.de/fit-connect/set-payload/2.2.3/set-payload.schema.json")));
    }

    @Test
    void isAllowedMetadataSchema() {
        assertTrue(underTest.isAllowedMetadataSchema(SchemaConfig.METADATA_V_1_0_0.getSchemaUri()));
    }

    @Test
    void isNoAllowedMetadataSchema() {
        assertFalse(underTest.isAllowedSetSchema(URI.create("https://schema.fitko.de/fit-connect/metadata/9.9.9/metadata.schema.json")));
    }

    @Test
    void loadSetSchema() throws JsonProcessingException {
        // Given
        final URI schemaUri = SchemaConfig.SET_V_1_0_1.getSchemaUri();

        // When
        final String setSchema = underTest.loadLatestSetSchema();

        //Then
        assertThat(new ObjectMapper().readTree(setSchema).get("$id").asText(), equalTo(schemaUri.toString()));
    }

    @Test
    void loadMetadataSchema() throws JsonProcessingException {
        // Given
        final URI schemaUri = SchemaConfig.METADATA_V_1_0_0.getSchemaUri();

        // When
        final String metadataSchema = underTest.loadMetadataSchema(schemaUri);

        //Then
        assertThat(new ObjectMapper().readTree(metadataSchema).get("$id").asText(), equalTo(schemaUri.toString()));
    }

    @Test
    void loadDestinationSchema() throws JsonProcessingException {
        // Given
        final URI schemaUri = SchemaConfig.XZUFI_DESTINATION_SCHEMA.getSchemaUri();

        // When
        final String metadataSchema = underTest.loadDestinationSchema(schemaUri);

        //Then
        assertThat(new ObjectMapper().readTree(metadataSchema).get("$id").asText(), equalTo(schemaUri.toString()));
    }

    @Test
    void loadMetadataSchemaThatDoesNotExist() {
        final URI schemaUri = URI.create("https://schema.fitko.de/fit-connect/metadata/9.9.9/metadata.schema.json");
        assertThrows(SchemaNotFoundException.class, () -> underTest.loadMetadataSchema(schemaUri));
    }
}