package dev.fitko.fitconnect.core.util;

import com.nimbusds.jwt.JWTClaimsSet;
import dev.fitko.fitconnect.api.domain.model.event.Event;
import dev.fitko.fitconnect.api.domain.model.event.EventIssuer;
import dev.fitko.fitconnect.api.domain.model.event.EventLog;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.problems.submission.InvalidEventLog;
import dev.fitko.fitconnect.api.exceptions.EventLogException;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import static dev.fitko.fitconnect.api.domain.model.event.Event.ACCEPT;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.*;

class EventLogUtilTest {

    @Test
    void mapEventLogToEntries() throws IOException {

        // Given
        final var setEvent = getResourceAsString("/signed_reject_event.txt");

        final EventLog eventLog = new EventLog();
        eventLog.setEventLogs(List.of(setEvent));

        // When
        final List<EventLogEntry> eventLogEntries = EventLogUtil.mapEventLogToEntries(eventLog);

        assertFalse(eventLogEntries.isEmpty());
        final EventLogEntry logEntry = eventLogEntries.get(0);

        // Then
        assertThat(logEntry.getEvent(), is(Event.REJECT));
        assertThat(logEntry.getIssuer(), is("1b7d1a24-a6c8-4050-bb71-ae3749ec432f"));
        assertThat(logEntry.getProblems().size(), is(1));
        assertThat(((Map)logEntry.getProblems().get(0)).get("type"), is(new InvalidEventLog().getType()));
    }

    @Test
    void resolveIssuerType() {

        // Given
        final var destinationIssuer = UUID.randomUUID().toString();
        final var submissionServiceIssuer = "https://submission-api.testing.fitko.dev";

        // Then
        assertThat(EventLogUtil.resolveIssuerType(submissionServiceIssuer), is(EventIssuer.SUBMISSION_SERVICE));
        assertThat(EventLogUtil.resolveIssuerType(destinationIssuer), is(EventIssuer.DESTINATION));
    }

    @Test
    void getEventType() {

        // Given
        final JWTClaimsSet claims = new JWTClaimsSet.Builder().claim("events", Map.of(Event.ACCEPT.getSchemaUri(), new Object())).build();

        // When
        final Event event = EventLogUtil.getEventFromClaims(claims);

        // Then
        assertThat(event, is(ACCEPT));

    }

    @Test
    void getIllegalEventType() {

        // Given
        final JWTClaimsSet claims = new JWTClaimsSet.Builder().claim("events", Map.of("http://foo-event", new Object())).build();

        // When
        final Event event = EventLogUtil.getEventFromClaims(claims);

        // Then
        assertNull(event);
    }

    @Test
    void getDestinationId() {

        // Given
        final var destinationIssuer = UUID.randomUUID();

        // When
        final UUID destinationId = EventLogUtil.getDestinationId(destinationIssuer.toString());

        // Then
        assertNotNull(destinationId);
        assertThat(destinationId, is(destinationIssuer));

    }

    @Test
    void getDestinationIdFromSubmissionServiceIssuer() {

        // Given
        final var submissionServiceIssuer =  "https://submission-api.testing.fitko.dev";

        // When
        final UUID destinationId = EventLogUtil.getDestinationId(submissionServiceIssuer);

        // Then
        assertNull(destinationId);

    }

    @Test
    void getMalformedDestinationIdFromDestinationIssuer() {

        // Given
        final var invalidUUID = "no-valid-uuid";

        // When
        final EventLogException exception = assertThrows(EventLogException.class, () -> EventLogUtil.getDestinationId(invalidUUID));

        // Then
        assertThat(exception.getMessage(), containsString("Destination id '" + invalidUUID + "' from issuer is no valid uuid"));

    }

    private String getResourceAsString(final String filename) throws IOException {
        return new String(EventLogUtilTest.class.getResourceAsStream(filename).readAllBytes());
    }
}